# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_utils import strutils

from futurist import periodics
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib.api.definitions import floating_ip_port_forwarding as pf_def
from neutron_lib.api.definitions import provider_net as provnet_apidef
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import constants as n_const
from neutron_lib import context as n_context
from neutron_lib.exceptions import l3 as lib_l3_exc
from oslo_utils import uuidutils

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf as ovn_config
from neutron.db import ovn_revision_numbers_db as db_rev
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.extensions import qos \
    as qos_extension
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import maintenance
from neutron.services.portforwarding import constants as pf_consts
from neutron.tests.functional import base
from neutron.tests.functional.services.logapi.drivers.ovn \
    import test_driver as test_log_driver

from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.extensions import test_extraroute


CFG_NEW_BURST = 50
CFG_NEW_RATE = 150


class _TestMaintenanceHelper(base.TestOVNFunctionalBase):
    """A helper class to keep the code more organized."""

    def setUp(self):
        super().setUp()
        self._ovn_client = self.mech_driver._ovn_client
        self._l3_ovn_client = self.l3_plugin._ovn_client
        ext_mgr = test_extraroute.ExtraRouteTestExtensionManager()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        self.maint = maintenance.DBInconsistenciesPeriodics(self._ovn_client)
        # Release the unneeded lock
        self.maint._idl.set_lock(None)
        self.context = n_context.get_admin_context()
        # Always verify inconsistencies for all objects.
        db_rev.INCONSISTENCIES_OLDER_THAN = -1

    def _find_network_row_by_name(self, name):
        for row in self.nb_api._tables['Logical_Switch'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY) == name):
                return row

    def _create_network(self, name, external=False, provider=None,
                        net_type=None):
        data = {'network': {'name': name,
                            extnet_apidef.EXTERNAL: external}}
        if net_type:
            data['network'][provnet_apidef.NETWORK_TYPE] = net_type
        if provider:
            net_type = net_type or 'flat'
            data['network'][provnet_apidef.NETWORK_TYPE] = net_type
            data['network'][provnet_apidef.PHYSICAL_NETWORK] = provider

        req = self.new_create_request('networks', data, self.fmt,
                                      as_admin=True)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['network']

    def _update_network_name(self, net_id, new_name):
        data = {'network': {'name': new_name}}
        req = self.new_update_request('networks', data, net_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['network']

    def _create_port(self, name, net_id, security_groups=None,
                     device_owner=None, vnic_type=None):
        data = {'port': {'name': name,
                         'network_id': net_id}}

        if security_groups is not None:
            data['port']['security_groups'] = security_groups

        if device_owner is not None:
            data['port']['device_owner'] = device_owner

        if vnic_type is not None:
            data['port']['binding:vnic_type'] = vnic_type

        req = self.new_create_request('ports', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['port']

    def _update_port_name(self, port_id, new_name):
        data = {'port': {'name': new_name}}
        req = self.new_update_request('ports', data, port_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['port']

    def _find_port_row_by_name(self, name):
        for row in self.nb_api._tables['Logical_Switch_Port'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_PORT_NAME_EXT_ID_KEY) == name):
                return row

    def _set_global_dhcp_opts(self, ip_version, opts):
        opt_string = ','.join([f'{key}:{value}'
                               for key, value
                               in opts.items()])
        if ip_version == n_const.IP_VERSION_6:
            ovn_config.cfg.CONF.set_override('ovn_dhcp6_global_options',
                                             opt_string,
                                             group='ovn')
        if ip_version == n_const.IP_VERSION_4:
            ovn_config.cfg.CONF.set_override('ovn_dhcp4_global_options',
                                             opt_string,
                                             group='ovn')

    def _unset_global_dhcp_opts(self, ip_version):
        if ip_version == n_const.IP_VERSION_6:
            ovn_config.cfg.CONF.clear_override('ovn_dhcp6_global_options',
                                               group='ovn')
        if ip_version == n_const.IP_VERSION_4:
            ovn_config.cfg.CONF.clear_override('ovn_dhcp4_global_options',
                                               group='ovn')

    def _create_subnet(self, name, net_id, ip_version=n_const.IP_VERSION_4,
                       **kwargs):
        if ip_version == n_const.IP_VERSION_4:
            cidr = '10.0.0.0/24'
        else:
            cidr = '2001:db8::/64'
        data = {'subnet': {'name': name,
                           'network_id': net_id,
                           'ip_version': ip_version,
                           'cidr': cidr,
                           'enable_dhcp': True}}
        data['subnet'].update(kwargs)
        req = self.new_create_request('subnets', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['subnet']

    def _update_subnet_enable_dhcp(self, subnet_id, value):
        data = {'subnet': {'enable_dhcp': value}}
        req = self.new_update_request('subnets', data, subnet_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['subnet']

    def _find_subnet_row_by_id(self, subnet_id):
        for row in self.nb_api._tables['DHCP_Options'].rows.values():
            if (row.external_ids.get('subnet_id') == subnet_id and
               not row.external_ids.get('port_id')):
                return row

    def _create_router(self, name, external_gateway_info=None):
        data = {'router': {'name': name}}
        as_admin = False
        if external_gateway_info is not None:
            data['router']['external_gateway_info'] = external_gateway_info
            as_admin = bool(external_gateway_info.get('enable_snat'))
        req = self.new_create_request('routers', data, self.fmt,
                                      as_admin=as_admin)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['router']

    def _update_router(self, router_id, router_dict):
        data = {'router': router_dict}
        req = self.new_update_request('routers', data, router_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['router']

    def _find_router_row_by_name(self, name):
        for row in self.nb_api._tables['Logical_Router'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY) == name):
                return row

    def _get_lrp_ext_ids_router_name(self, port_id):
        row = self._find_router_port_row_by_port_id(port_id)
        return row.external_ids.get(ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY)

    def _create_security_group(self):
        data = {'security_group': {'name': 'sgtest',
                                   'description': 'SpongeBob Rocks!'}}
        req = self.new_create_request('security-groups', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['security_group']

    def _find_security_group_row_by_id(self, sg_id):
        return self.nb_api.lookup(
            'Port_Group', utils.ovn_port_group_name(sg_id), default=None)

    def _create_security_group_rule(self, sg_id,
                                    remote_address_group_id=None,
                                    ethertype=n_const.IPv4):
        data = {'security_group_rule': {'security_group_id': sg_id,
                                        'direction': 'ingress',
                                        'protocol': n_const.PROTO_NAME_TCP,
                                        'ethertype': ethertype,
                                        'port_range_min': 22,
                                        'port_range_max': 22}}
        if remote_address_group_id:
            data['security_group_rule']['remote_address_group_id'] = (
                remote_address_group_id)
        req = self.new_create_request('security-group-rules', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['security_group_rule']

    def _create_address_group(self, name='ag_test',
                              addresses=None):
        if addresses is None:
            addresses = ['192.168.2.2/32', '2001:db8::/32']
        ag_args = {'project_id': self._project_id,
                   'name': name,
                   'description': 'test address group',
                   'addresses': addresses}
        return self._ovn_client._plugin.create_address_group(
            self.context, {'address_group': ag_args})

    def _find_address_set_for_ag(self, ag_id):
        _as = {}
        for ip_version in (n_const.IP_VERSION_4,
                           n_const.IP_VERSION_6):
            as_name = utils.ovn_ag_addrset_name(ag_id, 'ip' + str(ip_version))
            _as[ip_version] = self.nb_api.lookup(
                'Address_Set', as_name, default=None)
        return _as

    def _delete_address_set_for_ag(self, ag_id):
        """Delete OVN Address_Set rows for an address group."""
        for ip_version in n_const.IP_ALLOWED_VERSIONS:
            as_name = utils.ovn_ag_addrset_name(
                ag_id, 'ip' + str(ip_version))
            self.nb_api.address_set_del(
                as_name, if_exists=True).execute(check_error=True)

    def _find_security_group_rule_row_by_id(self, sgr_id):
        for row in self.nb_api._tables['ACL'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_SG_RULE_EXT_ID_KEY) == sgr_id):
                return row

    def _process_router_interface(self, action, router_id, subnet_id):
        req = self.new_action_request(
            'routers', {'subnet_id': subnet_id}, router_id,
            '%s_router_interface' % action)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)

    def _add_router_interface(self, router_id, subnet_id):
        return self._process_router_interface('add', router_id, subnet_id)

    def _remove_router_interface(self, router_id, subnet_id):
        return self._process_router_interface('remove', router_id, subnet_id)

    def _find_router_port_row_by_port_id(self, port_id):
        for row in self.nb_api._tables['Logical_Router_Port'].rows.values():
            if row.name == utils.ovn_lrouter_port_name(port_id):
                return row

    def _find_nat_rule(self, router_id, external_ip, logical_ip=None,
                       nat_type='dnat_and_snat'):
        rules = self.nb_api.get_lrouter_nat_rules(utils.ovn_name(router_id))
        return next((r for r in rules
                     if r['type'] == nat_type and
                     r['external_ip'] == external_ip and
                     (not logical_ip or r['logical_ip'] == logical_ip)),
                    None)

    def _find_pf_lb(self, router_id, fip_id=None):
        lbs = self.nb_api.get_router_floatingip_lbs(utils.ovn_name(router_id))
        return [lb for lb in lbs
                if (not fip_id or
                    fip_id == lb.external_ids[ovn_const.OVN_FIP_EXT_ID_KEY])]


class TestMaintenance(_TestMaintenanceHelper):

    def test_network(self):
        net_name = 'networktest'
        with mock.patch.object(self._ovn_client, 'create_network'):
            neutron_obj = self._create_network(net_name)

        # Assert the network doesn't exist in OVN
        self.assertIsNone(self._find_network_row_by_name(net_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the network was now created
        ovn_obj = self._find_network_row_by_name(net_name)
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        new_obj_name = 'networktest_updated'
        with mock.patch.object(self._ovn_client, 'update_network'):
            new_neutron_obj = self._update_network_name(neutron_obj['id'],
                                                        new_obj_name)

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_network_row_by_name(net_name)
        self.assertNotEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB
        self.assertIsNone(self._find_network_row_by_name(net_name))

        # Assert the network is now in sync
        ovn_obj = self._find_network_row_by_name(new_obj_name)
        self.assertEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_network'):
            self._delete('networks', new_neutron_obj['id'])

        # Assert the network still exists in OVNDB
        self.assertIsNotNone(self._find_network_row_by_name(new_obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the network is now deleted from OVNDB
        self.assertIsNone(self._find_network_row_by_name(new_obj_name))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            new_neutron_obj['id']))

    def test_port(self):
        obj_name = 'porttest'
        neutron_net = self._create_network('network1')

        with mock.patch.object(self._ovn_client, 'create_port'):
            neutron_obj = self._create_port(obj_name, neutron_net['id'])

        # Assert the port doesn't exist in OVN
        self.assertIsNone(self._find_port_row_by_name(obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the port was now created
        ovn_obj = self._find_port_row_by_name(obj_name)
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        new_obj_name = 'porttest_updated'
        with mock.patch.object(self._ovn_client, 'update_port'):
            new_neutron_obj = self._update_port_name(neutron_obj['id'],
                                                     new_obj_name)

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_port_row_by_name(obj_name)
        self.assertNotEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB
        self.assertIsNone(self._find_port_row_by_name(obj_name))

        # Assert the port is now in sync. Note that for ports we are
        # fetching it again from the Neutron database prior to comparison
        # because of the monitor code that can update the ports again upon
        # changes to it.
        ovn_obj = self._find_port_row_by_name(new_obj_name)
        new_neutron_obj = self._ovn_client._plugin.get_port(
            self.context, neutron_obj['id'])
        self.assertEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_port'):
            self._delete('ports', new_neutron_obj['id'])

        # Assert the port still exists in OVNDB
        self.assertIsNotNone(self._find_port_row_by_name(new_obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the port is now deleted from OVNDB
        self.assertIsNone(self._find_port_row_by_name(new_obj_name))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['id'],
            resource_type=ovn_const.TYPE_PORTS))

    def test_subnet_global_dhcp4_opts(self):
        obj_name = 'globaltestsubnet'
        options = {'ntp_server': '1.2.3.4'}
        neutron_net = self._create_network('network1')

        # Create a subnet without global options
        neutron_sub = self._create_subnet(obj_name, neutron_net['id'])

        # Assert that the option is not set
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

        # Set some global DHCP Options
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_4,
                                   opts=options)

        # Run the maintenance task to add the new options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was added
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '1.2.3.4',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option
        new_options = {'ntp_server': '4.3.2.1'}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_4,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was changed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '4.3.2.1',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option to null
        new_options = {'ntp_server': ''}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_4,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was removed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

    def test_subnet_global_dhcp6_opts(self):
        obj_name = 'globaltestsubnet'
        options = {'ntp_server': '1.2.3.4'}
        neutron_net = self._create_network('network1')

        # Create a subnet without global options
        neutron_sub = self._create_subnet(obj_name, neutron_net['id'],
                                          n_const.IP_VERSION_6)

        # Assert that the option is not set
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

        # Set some global DHCP Options
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_6,
                                   opts=options)

        # Run the maintenance task to add the new options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was added
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '1.2.3.4',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option
        new_options = {'ntp_server': '4.3.2.1'}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_6,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was changed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '4.3.2.1',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option to null
        new_options = {'ntp_server': ''}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_6,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was removed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

    def test_check_dhcp_options_consistency(self):
        neutron_net = self._create_network('dhcp_consistency_net')
        neutron_sub = self._create_subnet('dhcp_consistency_subnet',
                                          neutron_net['id'])

        dhcp_row = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNotNone(dhcp_row)
        self.assertEqual(
            neutron_net['id'],
            dhcp_row.external_ids.get(ovn_const.OVN_NETWORK_ID_EXT_ID_KEY))

        # Simulate legacy row: remove network_id from external_ids
        self.nb_api.db_remove(
            'DHCP_Options', dhcp_row.uuid, 'external_ids',
            ovn_const.OVN_NETWORK_ID_EXT_ID_KEY).execute(check_error=True)

        dhcp_row = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertNotIn(ovn_const.OVN_NETWORK_ID_EXT_ID_KEY,
                         dhcp_row.external_ids)

        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_dhcp_options_consistency)

        dhcp_row = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            neutron_net['id'],
            dhcp_row.external_ids.get(ovn_const.OVN_NETWORK_ID_EXT_ID_KEY))

        # Test original external_ids are not modified
        self.assertEqual(neutron_sub['id'], dhcp_row.external_ids['subnet_id'])

    def test_subnet(self):
        obj_name = 'subnettest'
        neutron_net = self._create_network('network1')

        with mock.patch.object(self._ovn_client, 'create_subnet'):
            neutron_obj = self._create_subnet(obj_name, neutron_net['id'])

        # Assert the subnet doesn't exist in OVN
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the subnet was now created
        ovn_obj = self._find_subnet_row_by_id(neutron_obj['id'])
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        with mock.patch.object(self._ovn_client, 'update_subnet'):
            neutron_obj = self._update_subnet_enable_dhcp(
                neutron_obj['id'], False)

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_subnet_row_by_id(neutron_obj['id'])
        self.assertNotEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB. When
        # the subnet's enable_dhcp's is set to False, OVN will remove the
        # DHCP_Options entry related to that subnet.
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Re-enable the DHCP for the subnet and check if the maintenance
        # thread will re-create it in OVN
        with mock.patch.object(self._ovn_client, 'update_subnet'):
            neutron_obj = self._update_subnet_enable_dhcp(
                neutron_obj['id'], True)

        # Assert the DHCP_Options still doesn't exist in OVNDB
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the subnet is now in sync
        ovn_obj = self._find_subnet_row_by_id(neutron_obj['id'])
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_subnet'):
            self._delete('subnets', neutron_obj['id'])

        # Assert the subnet still exists in OVNDB
        self.assertIsNotNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the subnet is now deleted from OVNDB
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['id']))

    def test_router(self):
        obj_name = 'routertest'

        with mock.patch.object(self._l3_ovn_client, 'create_router'):
            neutron_obj = self._create_router(obj_name)

        # Assert the router doesn't exist in OVN
        self.assertIsNone(self._find_router_row_by_name(obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router was now created
        ovn_obj = self._find_router_row_by_name(obj_name)
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        new_obj_name = 'routertest_updated'
        with mock.patch.object(self._l3_ovn_client, 'update_router'):
            new_neutron_obj = self._update_router(neutron_obj['id'],
                                                  {'name': new_obj_name})

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_router_row_by_name(obj_name)
        self.assertNotEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB
        self.assertIsNone(self._find_router_row_by_name(obj_name))

        # Assert the router is now in sync
        ovn_obj = self._find_router_row_by_name(new_obj_name)
        self.assertEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._l3_ovn_client, 'delete_router'):
            self._delete('routers', new_neutron_obj['id'])

        # Assert the router still exists in OVNDB
        self.assertIsNotNone(self._find_router_row_by_name(new_obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router is now deleted from OVNDB
        self.assertIsNone(self._find_router_row_by_name(new_obj_name))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            new_neutron_obj['id']))

    def test_security_group(self):
        with mock.patch.object(self._ovn_client, 'create_security_group'):
            neutron_obj = self._create_security_group()

        # Assert the sg doesn't exist in OVN
        self.assertIsNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the sg was now created. We don't save the revision number
        # in the Security Group because OVN doesn't support updating it,
        # all we care about is whether it exists or not.
        self.assertIsNotNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_security_group'):
            self._delete('security-groups', neutron_obj['id'])

        # Assert the sg still exists in OVNDB
        self.assertIsNotNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the sg is now deleted from OVNDB
        self.assertIsNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['id']))

    def test_security_group_rule(self):
        neutron_sg = self._create_security_group()
        neutron_net = self._create_network('network1')
        self._create_port('portsgtest', neutron_net['id'],
                          security_groups=[neutron_sg['id']])

        with mock.patch.object(self._ovn_client, 'create_security_group_rule'):
            neutron_obj = self._create_security_group_rule(neutron_sg['id'])

        # Assert the sg rule doesn't exist in OVN
        self.assertIsNone(
            self._find_security_group_rule_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the sg rule was now created. We don't save the revision number
        # in the Security Group because OVN doesn't support updating it,
        # all we care about is whether it exists or not.
        self.assertIsNotNone(
            self._find_security_group_rule_row_by_id(neutron_obj['id']))

        # > Delete

        # FIXME(lucasagomes): Maintenance thread fixing deleted
        # security group rules is currently broken due to:
        # https://bugs.launchpad.net/networking-ovn/+bug/1756123

    def test_router_port(self):
        neutron_net = self._create_network('networktest', external=True)
        neutron_subnet = self._create_subnet('subnettest', neutron_net['id'])
        neutron_router = self._create_router('routertest')

        with mock.patch.object(self._l3_ovn_client, 'create_router_port'):
            with mock.patch('neutron.db.ovn_revision_numbers_db.'
                            'bump_revision'):
                neutron_obj = self._add_router_interface(neutron_router['id'],
                                                         neutron_subnet['id'])

        # Assert the router port doesn't exist in OVN
        self.assertIsNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router port was now created
        self.assertIsNotNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # Assert router port has "neutron-" prefix and correct router uuid
        # for external_ids neutron:router_name value, when created and after
        # maintenance/updated (LP#2055045)

        self.assertEqual(
            '%s%s' % (ovn_const.OVN_NAME_PREFIX, neutron_router['id']),
            self._get_lrp_ext_ids_router_name(neutron_obj['port_id']))

        self.assertRaises(periodics.NeverAgain,
                          self.maint.update_lrouter_ports_ext_ids_name_prefix)

        self.assertEqual(
            '%s%s' % (ovn_const.OVN_NAME_PREFIX, neutron_router['id']),
            self._get_lrp_ext_ids_router_name(neutron_obj['port_id']))

        # > Delete

        with mock.patch.object(self._l3_ovn_client, 'delete_router_port'):
            self._remove_router_interface(neutron_router['id'],
                                          neutron_subnet['id'])

        # Assert the router port still exists in OVNDB
        self.assertIsNotNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router port is now deleted from OVNDB
        self.assertIsNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['port_id']))

    def test_check_for_igmp_snooping_enabled(self):
        cfg.CONF.set_override('igmp_snooping_enable', False, group='OVS')
        net = self._create_network('net')
        ls = self.nb_api.db_find(
            'Logical_Switch',
            ('name', '=', utils.ovn_name(net['id']))).execute(
                check_error=True)[0]

        self.assertEqual('false', ls['other_config'][ovn_const.MCAST_SNOOP])
        self.assertEqual(
            'false', ls['other_config'][ovn_const.MCAST_FLOOD_UNREGISTERED])

        # Change the value of the configuration
        cfg.CONF.set_override('igmp_snooping_enable', True, group='OVS')

        # Call the maintenance task and check that the value has been
        # updated in the Logical Switch
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_for_igmp_snoop_support)

        ls = self.nb_api.db_find(
            'Logical_Switch',
            ('name', '=', utils.ovn_name(net['id']))).execute(
                check_error=True)[0]

        self.assertEqual('true', ls['other_config'][ovn_const.MCAST_SNOOP])
        self.assertEqual(
            'false', ls['other_config'][ovn_const.MCAST_FLOOD_UNREGISTERED])

    def test_check_for_aging_settings(self):
        net = self._create_network('net', provider='datacentre')
        ls = self.nb_api.get_lswitch(utils.ovn_name(net['id']))

        self.assertEqual(
            str(ovn_const.FDB_AGE_THRESHOLD_DEFAULT),
            ls.other_config.get(ovn_const.LS_OPTIONS_FDB_AGE_THRESHOLD))

        self.assertEqual(
            '0', self.nb_api.nb_global.options.get("fdb_removal_limit", '0'))

        # Change the value of the configuration
        cfg.CONF.set_override('fdb_age_threshold', 5, group='ovn')
        cfg.CONF.set_override('fdb_removal_limit', 100, group='ovn_nb_global')

        # Call the maintenance task and check that the value has been
        # updated in the Logical Switch
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_fdb_aging_settings)

        ls = self.nb_api.get_lswitch(utils.ovn_name(net['id']))

        self.assertEqual(
            '5', ls.other_config.get(ovn_const.LS_OPTIONS_FDB_AGE_THRESHOLD))
        self.assertEqual(
            '100', self.nb_api.nb_global.options.get("fdb_removal_limit"))

    def test_update_mac_aging_settings(self):
        ext_net = self._create_network('ext_networktest', external=True)
        ext_subnet = self._create_subnet(
            'ext_subnettest',
            ext_net['id'],
            **{'cidr': '100.0.0.0/24',
               'gateway_ip': '100.0.0.254',
               'allocation_pools': [
                   {'start': '100.0.0.2', 'end': '100.0.0.253'}],
               'enable_dhcp': False})
        self._create_network('network1test', external=False)
        external_gateway_info = {
            'enable_snat': True,
            'network_id': ext_net['id'],
            'external_fixed_ips': [
                {'ip_address': '100.0.0.2', 'subnet_id': ext_subnet['id']}]}
        router = self._create_router(
            'routertest', external_gateway_info=external_gateway_info)

        options = self.nb_api.nb_global.options
        lr = self.nb_api.get_lrouter(router["id"])

        self.assertEqual(
            str(ovn_const.MAC_BINDING_AGE_THRESHOLD),
            lr.options.get(ovn_const.LR_OPTIONS_MAC_AGE_LIMIT))

        self.assertEqual('0', options.get('mac_binding_removal_limit', '0'))

        cfg.CONF.set_override("mac_binding_age_threshold", 5, group="ovn")
        cfg.CONF.set_override("mac_binding_removal_limit", 100,
                              group="ovn_nb_global")

        # Call the maintenance task and check that the value has been
        # updated in the Logical Switch
        self.assertRaises(periodics.NeverAgain,
                          self.maint.update_mac_aging_settings)

        lr = self.nb_api.get_lrouter(router['id'])
        options = self.nb_api.nb_global.options

        self.assertEqual(
            '5', lr.options.get(ovn_const.LR_OPTIONS_MAC_AGE_LIMIT))
        self.assertEqual('100', options['mac_binding_removal_limit'])

    def test_floating_ip(self):
        ext_net = self._create_network('ext_networktest', external=True)
        ext_subnet = self._create_subnet(
            'ext_subnettest',
            ext_net['id'],
            **{'cidr': '100.0.0.0/24',
               'gateway_ip': '100.0.0.254',
               'allocation_pools': [
                   {'start': '100.0.0.2', 'end': '100.0.0.253'}],
               'enable_dhcp': False})
        net1 = self._create_network('network1test', external=False)
        subnet1 = self._create_subnet('subnet1test', net1['id'])
        external_gateway_info = {
            'enable_snat': True,
            'network_id': ext_net['id'],
            'external_fixed_ips': [
                {'ip_address': '100.0.0.2', 'subnet_id': ext_subnet['id']}]}
        router = self._create_router(
            'routertest', external_gateway_info=external_gateway_info)
        self._add_router_interface(router['id'], subnet1['id'])

        p1 = self._create_port('testp1', net1['id'])
        logical_ip = p1['fixed_ips'][0]['ip_address']
        fip_info = {'floatingip': {
            'project_id': self._project_id,
            'description': 'test_fip',
            'floating_network_id': ext_net['id'],
            'port_id': p1['id'],
            'fixed_ip_address': logical_ip}}

        # > Create
        with mock.patch.object(self._l3_ovn_client, 'create_floatingip'):
            fip = self.l3_plugin.create_floatingip(self.context, fip_info)

        floating_ip_address = fip['floating_ip_address']
        self.assertEqual(router['id'], fip['router_id'])
        self.assertEqual('testp1', fip['port_details']['name'])
        self.assertIsNotNone(self.nb_api.get_lswitch_port(fip['port_id']))

        # Assert the dnat_and_snat rule doesn't exist in OVN
        self.assertIsNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the rule for the fip is now present
        self.assertIsNotNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # > Update
        p2 = self._create_port('testp2', net1['id'])
        logical_ip = p2['fixed_ips'][0]['ip_address']
        fip_info = {'floatingip': {
            'port_id': p2['id'],
            'fixed_ip_address': logical_ip}}

        with mock.patch.object(self._l3_ovn_client, 'update_floatingip'):
            self.l3_plugin.update_floatingip(self.context, fip['id'], fip_info)

        # Assert the dnat_and_snat rule in OVN is still using p1's address
        stale_nat_rule = self._find_nat_rule(router['id'], floating_ip_address)
        self.assertEqual(p1['fixed_ips'][0]['ip_address'],
                         stale_nat_rule['logical_ip'])

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the rule for the fip is now updated
        self.assertIsNotNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # > Delete
        with mock.patch.object(self._l3_ovn_client, 'delete_floatingip'):
            self.l3_plugin.delete_floatingip(self.context, fip['id'])

        self.assertRaises(
            lib_l3_exc.FloatingIPNotFound,
            self.l3_plugin.get_floatingip, self.context, fip['id'])

        # Assert the dnat_and_snat rule in OVN is still present
        self.assertIsNotNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the rule for the fip is now gone
        self.assertIsNone(
            self._find_nat_rule(router['id'], floating_ip_address))

        # Assert the router snat rule is still there
        snat_rule = self._find_nat_rule(
            router['id'], '100.0.0.2', nat_type='snat')
        self.assertEqual(subnet1['cidr'], snat_rule['logical_ip'])

    def test_port_forwarding(self):
        def fip_attrs(args):
            return {pf_def.RESOURCE_NAME: {pf_def.RESOURCE_NAME: args}}

        def _verify_lb(test, protocol, vip_ext_port, vip_int_port):
            ovn_lbs = self._find_pf_lb(router_id, fip_id)
            test.assertEqual(1, len(ovn_lbs))
            test.assertEqual(f'pf-floatingip-{fip_id}-{protocol}',
                             ovn_lbs[0].name)
            test.assertEqual(
                {f'{fip_ip}:{vip_ext_port}':
                 f'{p1_ip}:{vip_int_port}'},
                ovn_lbs[0].vips)

        ext_net = self._create_network('ext_networktest', external=True)
        ext_subnet = self._create_subnet(
            'ext_subnettest',
            ext_net['id'],
            **{'cidr': '100.0.0.0/24',
               'gateway_ip': '100.0.0.254',
               'allocation_pools': [
                   {'start': '100.0.0.2', 'end': '100.0.0.253'}],
               'enable_dhcp': False})
        net1 = self._create_network('network1test', external=False)
        subnet1 = self._create_subnet('subnet1test', net1['id'])
        external_gateway_info = {
            'enable_snat': True,
            'network_id': ext_net['id'],
            'external_fixed_ips': [
                {'ip_address': '100.0.0.2', 'subnet_id': ext_subnet['id']}]}
        router = self._create_router(
            'routertest', external_gateway_info=external_gateway_info)
        router_id = router['id']
        self._add_router_interface(router['id'], subnet1['id'])

        fip_info = {'floatingip': {
            'project_id': self._project_id,
            'floating_network_id': ext_net['id'],
            'port_id': None,
            'fixed_ip_address': None}}
        fip = self.l3_plugin.create_floatingip(self.context, fip_info)
        fip_id = fip['id']
        fip_ip = fip['floating_ip_address']
        p1 = self._create_port('testp1', net1['id'])
        p1_ip = p1['fixed_ips'][0]['ip_address']

        callbacks = registry._get_callback_manager()._callbacks
        pf_cb = callbacks[pf_consts.PORT_FORWARDING]
        key = list(pf_cb[events.AFTER_UPDATE][0][1].keys())[0]
        pf_cb[events.AFTER_CREATE][0][1][key] = mock.MagicMock()

        # > Create
        fip_pf_args = {
            pf_def.EXTERNAL_PORT: 2222,
            pf_def.INTERNAL_PORT: 22,
            pf_def.INTERNAL_PORT_ID: p1['id'],
            pf_def.PROTOCOL: 'tcp',
            pf_def.INTERNAL_IP_ADDRESS: p1_ip}
        pf_obj = self.pf_plugin.create_floatingip_port_forwarding(
            self.context, fip_id, **fip_attrs(fip_pf_args))
        call = mock.call('port_forwarding', 'after_create', self.pf_plugin,
                         payload=mock.ANY)
        pf_cb[events.AFTER_CREATE][0][1][key].assert_has_calls([call])

        # Assert load balancer for port forwarding was not created
        self.assertFalse(self._find_pf_lb(router_id, fip_id))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert load balancer for port forwarding was created
        _verify_lb(self, 'tcp', 2222, 22)

        # > Update
        fip_pf_args = {pf_def.EXTERNAL_PORT: 5353,
                       pf_def.INTERNAL_PORT: 53,
                       pf_def.PROTOCOL: 'udp'}
        pf_cb[events.AFTER_UPDATE][0][1][key] = mock.MagicMock()
        self.pf_plugin.update_floatingip_port_forwarding(
            self.context, pf_obj['id'], fip_id, **fip_attrs(fip_pf_args))
        call = mock.call('port_forwarding', 'after_update', self.pf_plugin,
                         payload=mock.ANY)
        pf_cb[events.AFTER_UPDATE][0][1][key].assert_has_calls([call])

        # Assert load balancer for port forwarding is stale
        _verify_lb(self, 'tcp', 2222, 22)

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert load balancer for port forwarding was updated
        _verify_lb(self, 'udp', 5353, 53)

        # > Delete
        pf_cb[events.AFTER_DELETE][0][1][key] = mock.MagicMock()
        self.pf_plugin.delete_floatingip_port_forwarding(
            self.context, pf_obj['id'], fip_id)
        call = mock.call('port_forwarding', 'after_delete', self.pf_plugin,
                         payload=mock.ANY)
        pf_cb[events.AFTER_DELETE][0][1][key].assert_has_calls([call])

        # Assert load balancer for port forwarding is stale
        _verify_lb(self, 'udp', 5353, 53)

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert load balancer for port forwarding is gone
        self.assertFalse(self._find_pf_lb(router_id, fip_id))

    def _test_check_provider_distributed_ports(
            self, is_distributed_fip, net_type, expected_value=None):
        cfg.CONF.set_override(
            'enable_distributed_floating_ip', is_distributed_fip, group='ovn')
        net_args = {'net_type': net_type}
        if net_type == n_const.TYPE_FLAT:
            net_args['provider'] = 'datacentre'
        net = self._create_network(
            'net_distributed_ports_test', **net_args)
        subnet = self._create_subnet('subnet_distributed_ports_test',
                                     net['id'])
        router = self._create_router('router_distributed_ports_test')
        self._add_router_interface(router['id'], subnet['id'])

        # Lets make sure that reside-on-chassis-redirect is not set for the LRP
        lr = self.nb_api.lookup('Logical_Router',
                                utils.ovn_name(router['id']))
        lrp = lr.ports[0]
        self.nb_api.db_remove(
            'Logical_Router_Port',
            lrp.name,
            'options',
            ovn_const.LRP_OPTIONS_RESIDE_REDIR_CH
        ).execute()

        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_provider_distributed_ports)

        lrp = self.nb_api.lookup('Logical_Router_Port', lrp.name)
        if net_type in [n_const.TYPE_VLAN, n_const.TYPE_FLAT]:
            self.assertEqual(
                expected_value,
                strutils.bool_from_string(
                    lrp.options[ovn_const.LRP_OPTIONS_RESIDE_REDIR_CH]))
        else:
            self.assertNotIn(
                ovn_const.LRP_OPTIONS_RESIDE_REDIR_CH,
                lrp.options)

    def test_check_provider_distributed_ports_dvr_vlan_net(self):
        self._test_check_provider_distributed_ports(True, 'vlan', False)

    def test_check_provider_distributed_ports_non_dvr_vlan_net(self):
        self._test_check_provider_distributed_ports(False, 'vlan', True)

    def test_check_provider_distributed_ports_dvr_flat_net(self):
        self._test_check_provider_distributed_ports(True, 'flat', False)

    def test_check_provider_distributed_ports_non_dvr_flat_net(self):
        self._test_check_provider_distributed_ports(False, 'flat', True)

    def test_check_provider_distributed_ports_dvr_geneve_net(self):
        self._test_check_provider_distributed_ports(True, 'geneve')

    def test_check_provider_distributed_ports_non_dvr_geneve_net(self):
        self._test_check_provider_distributed_ports(False, 'geneve')

    def test_configure_nb_global(self):
        def options_intersect(options1, options2):
            return bool(set(
                new_nb_global_options.keys()).intersection(nb_options.keys()))

        new_nb_global_options = {
            'foo': 'bar',
            'baz': 'qux',
        }

        cfg_nb_global_options = [
            ovn_config.cfg.StrOpt(key) for key in new_nb_global_options
        ]

        def get_opt(key):
            return new_nb_global_options[key]

        nb_options = self.nb_api.db_get('NB_Global', '.', 'options').execute(
            check_error=True, log_errors=True)
        self.assertFalse(options_intersect(new_nb_global_options, nb_options))

        with mock.patch.object(
                ovn_config, 'nb_global_opts', cfg_nb_global_options), \
                mock.patch.object(
                    ovn_config.cfg.CONF.ovn_nb_global, 'get',
                    side_effect=get_opt):
            self.assertRaises(periodics.NeverAgain,
                              self.maint.configure_nb_global)

        nb_options = self.nb_api.db_get('NB_Global', '.', 'options').execute(
            check_error=True, log_errors=True)
        self.assertTrue(options_intersect(new_nb_global_options, nb_options))

    def test_set_network_type_and_physnet(self):
        net1 = self._create_network(uuidutils.generate_uuid())
        net2 = self._create_network(uuidutils.generate_uuid(),
                                    provider='physnet1', net_type='vlan')
        ls1_name = utils.ovn_name(net1['id'])
        ls2_name = utils.ovn_name(net2['id'])
        for _ls_name in (ls1_name, ls2_name):
            self.nb_api.db_remove(
                'Logical_Switch', _ls_name, 'external_ids',
                ovn_const.OVN_NETTYPE_EXT_ID_KEY).execute(check_error=True)
            self.nb_api.db_remove(
                'Logical_Switch', _ls_name, 'external_ids',
                ovn_const.OVN_PHYSNET_EXT_ID_KEY).execute(check_error=True)
            ls = self.nb_api.lookup('Logical_Switch', _ls_name)
            self.assertIsNone(ls.external_ids.get(
                ovn_const.OVN_NETTYPE_EXT_ID_KEY))
            self.assertIsNone(ls.external_ids.get(
                ovn_const.OVN_PHYSNET_EXT_ID_KEY))

        self.assertRaises(periodics.NeverAgain,
                          self.maint.set_network_type_and_physnet)
        ls1 = self.nb_api.lookup('Logical_Switch', ls1_name)
        self.assertEqual(
            net1[provnet_apidef.NETWORK_TYPE],
            ls1.external_ids.get(ovn_const.OVN_NETTYPE_EXT_ID_KEY))
        self.assertNotIn(
            ovn_const.OVN_PHYSNET_EXT_ID_KEY,
            ls1.external_ids.get)

        ls2 = self.nb_api.lookup('Logical_Switch', ls2_name)
        self.assertEqual(
            net2[provnet_apidef.NETWORK_TYPE],
            ls2.external_ids.get(ovn_const.OVN_NETTYPE_EXT_ID_KEY))
        self.assertEqual(
            net2[provnet_apidef.PHYSICAL_NETWORK],
            ls2.external_ids.get(ovn_const.OVN_PHYSNET_EXT_ID_KEY))

    def test_check_network_broadcast_arps_to_all_routers(self):
        net = self._create_network('net', external=True)
        ls = self.nb_api.get_lswitch(utils.ovn_name(net['id']))

        self.assertEqual(
            'true',
            ls.other_config.get(ovn_const.LS_OPTIONS_BROADCAST_ARPS_ROUTERS))

        # Change the value of the configuration
        cfg.CONF.set_override(
            'broadcast_arps_to_all_routers', False, group='ovn')

        # Call the maintenance task and check that the value has been
        # updated in the Logical Switch
        self.assertRaises(
            periodics.NeverAgain,
            self.maint.check_network_broadcast_arps_to_all_routers)

        ls = self.nb_api.get_lswitch(utils.ovn_name(net['id']))

        self.assertEqual(
            'false',
            ls.other_config.get(ovn_const.LS_OPTIONS_BROADCAST_ARPS_ROUTERS))

    def test_static_routes_with_external_ids(self):
        ext_net = self._create_network('ext_networktest', external=True)
        ext_subnet = self._create_subnet(
            'ext_subnettest',
            ext_net['id'],
            **{'cidr': '100.0.0.0/24',
               'gateway_ip': '100.0.0.254',
               'allocation_pools': [
                   {'start': '100.0.0.2', 'end': '100.0.0.253'}],
               'enable_dhcp': False})
        net1 = self._create_network('network1test', external=False)
        subnet1 = self._create_subnet('subnet1test', net1['id'])
        external_gateway_info = {
            'enable_snat': True,
            'network_id': ext_net['id'],
            'external_fixed_ips': [
                {'ip_address': '100.0.0.2', 'subnet_id': ext_subnet['id']}]}
        router = self._create_router(
            'routertest', external_gateway_info=external_gateway_info)
        self._add_router_interface(router['id'], subnet1['id'])

        # Create static routes via Neutron
        with mock.patch.object(self.nb_api,
                               'add_static_route', columns=None):
            self.l3_plugin.update_router(
                self.context, router['id'],
                {'router': {'routes': [{'destination': '10.10.0.0/24',
                                        'nexthop': '100.0.0.3'},
                                       {'destination': '20.0.0.0/24',
                                        'nexthop': '100.0.0.6'}]}})

        # Create a Neutron owned static route with external_ids key
        columns = {'external_ids': {ovn_const.OVN_LRSR_EXT_ID_KEY: 'true'}}
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.add_static_route('neutron-' + router['id'],
                                                 ip_prefix='10.10.0.0/24',
                                                 nexthop='100.0.0.3',
                                                 **columns))

        # Create a Neutron owned static route without external_ids key
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.add_static_route('neutron-' + router['id'],
                                                 ip_prefix='20.0.0.0/24',
                                                 nexthop='100.0.0.6'))

        # Create an OVN externally managed static route without external_ids
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.add_static_route('neutron-' + router['id'],
                                                 ip_prefix='30.0.0.0/24',
                                                 nexthop='100.0.0.9'))

        sroutes = self.nb_api.get_all_logical_routers_static_routes()[0]
        sroute_info = sroutes['static_routes']
        for route in sroute_info:
            if route.ip_prefix == '10.10.0.0/24':
                self.assertEqual(route.external_ids,
                                 {ovn_const.OVN_LRSR_EXT_ID_KEY: 'true'})
            if route.ip_prefix == '20.0.0.0/24':
                self.assertEqual({}, route.external_ids)
            if route.ip_prefix == '30.0.0.0/24':
                self.assertEqual({}, route.external_ids)

        # Call the maintenance task and check that the value has been
        # updated in the external_ids.
        self.assertRaises(periodics.NeverAgain,
                          self.maint.update_router_static_routes)

        sroutes = self.nb_api.get_all_logical_routers_static_routes()[0]
        sroute_info = sroutes['static_routes']
        for route in sroute_info:
            if route.ip_prefix == '10.10.0.0/24':
                self.assertEqual(route.external_ids,
                                 {ovn_const.OVN_LRSR_EXT_ID_KEY: 'true'})
            # Check if the OVN static route was updated with the Neutron key
            if route.ip_prefix == '20.0.0.0/24':
                self.assertEqual(route.external_ids,
                                 {ovn_const.OVN_LRSR_EXT_ID_KEY: 'true'})
            # Check if the externally managed OVN static route remains
            # without the Neutron key.
            if route.ip_prefix == '30.0.0.0/24':
                self.assertEqual({}, route.external_ids)

    def _get_nb_global_external_ids(self):
        return self.nb_api.db_get(
            'NB_Global', '.', 'external_ids').execute(check_error=True)

    def test_set_fip_distributed_flag(self):
        ovn_config.cfg.CONF.set_override(
            'enable_distributed_floating_ip', True, 'ovn')
        nb_global_ext_id = self._get_nb_global_external_ids()
        self.assertNotIn(ovn_const.OVN_FIP_DISTRIBUTED_KEY, nb_global_ext_id)

        self.assertRaises(
            periodics.NeverAgain, self.maint.set_fip_distributed_flag)

        nb_global_ext_id = self._get_nb_global_external_ids()
        self.assertEqual(
            "True", nb_global_ext_id[ovn_const.OVN_FIP_DISTRIBUTED_KEY])

    def _test_set_fip_distributed_flag_change(
            self, original_value, config_value):
        ovn_config.cfg.CONF.set_override(
            'enable_distributed_floating_ip', config_value, 'ovn')
        self.nb_api.db_set(
            'NB_Global', '.', external_ids={
                ovn_const.OVN_FIP_DISTRIBUTED_KEY: str(original_value)}
        ).execute(check_error=True)
        nb_global_ext_id = self._get_nb_global_external_ids()
        self.assertEqual(
            str(original_value),
            nb_global_ext_id[ovn_const.OVN_FIP_DISTRIBUTED_KEY])

        self.assertRaises(
            periodics.NeverAgain, self.maint.set_fip_distributed_flag)

        nb_global_ext_id = self._get_nb_global_external_ids()
        self.assertEqual(
            str(config_value),
            nb_global_ext_id[ovn_const.OVN_FIP_DISTRIBUTED_KEY])

    def test_set_fip_distributed_flag_changed(self):
        self._test_set_fip_distributed_flag_change(
            original_value=False,
            config_value=True)

    def test_set_fip_distributed_flag_unchanged(self):
        self._test_set_fip_distributed_flag_change(
            original_value=True,
            config_value=True)

    def test_set_ovn_owned_dns_option(self):
        neutron_net = self._create_network('network1')
        ls_name = utils.ovn_name(neutron_net['id'])
        with mock.patch.object(
                self._ovn_client, 'is_dns_required_for_port',
                return_value=True):
            self._create_port('portdns', neutron_net['id'])

        ls, ls_dns_record = self.nb_api.get_ls_and_dns_record(ls_name)

        # Assert that option is not set
        self.assertNotEqual(
            ls_dns_record.options.get('ovn-owned'), 'true')

        # Override config
        cfg.CONF.set_override(
            'dns_records_ovn_owned', True, group='ovn')

        # Call the maintenance task and check that the option has been
        # updated in the DNS record
        self.assertRaises(
            periodics.NeverAgain,
            self.maint.set_ovn_owned_dns_option)

        # Assert that option is not set
        self.assertEqual(
            ls_dns_record.options.get('ovn-owned'), 'true')

    def test_update_qos_fip_rule_priority(self):
        def_prio = qos_extension.OVN_QOS_DEFAULT_RULE_PRIORITY
        fip_prio = qos_extension.OVN_QOS_FIP_RULE_PRIORITY
        neutron_net = self._create_network('network1')
        ls_name = utils.ovn_name(neutron_net['id'])
        self.nb_api.qos_add(
            ls_name, 'from-lport', def_prio, "outport == 1",
            1000, 800, None, None,
            external_ids={ovn_const.OVN_ROUTER_ID_EXT_ID_KEY: 1})
        self.nb_api.qos_add(
            ls_name, 'from-lport', def_prio, "outport == 1",
            1000, 800, None, None,
            external_ids={ovn_const.OVN_FIP_EXT_ID_KEY: 1})

        self.assertRaises(
            periodics.NeverAgain,
            self.maint.update_qos_fip_rule_priority)

        for qos_rule in self.nb_api.qos_list(ls_name).execute(
                check_error=True):
            if qos_rule.external_ids.get(ovn_const.OVN_FIP_EXT_ID_KEY):
                self.assertEqual(fip_prio, qos_rule.priority)
            else:
                self.assertEqual(def_prio, qos_rule.priority)

    def test_update_lrouter_ports_ext_ids_name_prefix(self):
        router = self._create_router(uuidutils.generate_uuid())
        port_ids = []
        for idx in range(2):
            # Create 2 networks, 2 subnets and 2 router interfaces.
            net = self._create_network(uuidutils.generate_uuid())
            subnet = self._create_subnet(uuidutils.generate_uuid(), net['id'],
                                         cidr=f'10.{idx}.0.0/24')
            port_ids.append(self._add_router_interface(
                router['id'], subnet['id'])['port_id'])

        # Remove the 'OVN_ROUTER_NAME_EXT_ID_KEY' from the second LRP.
        self.nb_api.db_remove(
            'Logical_Router_Port', 'lrp-' + port_ids[1],
            'external_ids', ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY).execute(
                check_error=True)
        lrp = self.nb_api.db_find(
            'Logical_Router_Port',
            ('name', '=', 'lrp-' + port_ids[1])).execute(check_error=True)[0]
        self.assertNotIn(ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY,
                         lrp['external_ids'])

        # Set the router ID (not the OVN router name) in the first LRP.
        # This one will be updated by the maintenance method.
        router_name = {
            ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: router['id']}
        self.nb_api.db_set(
            'Logical_Router_Port', 'lrp-' + port_ids[0],
            ('external_ids', router_name)).execute(check_error=True)

        self.assertRaises(
            periodics.NeverAgain,
            self.maint.update_lrouter_ports_ext_ids_name_prefix)

        lrps = self.nb_api.db_find(
            'Logical_Router_Port').execute(check_error=True)
        for lrp in (lrp for lrp in lrps if
                    lrp['name'] == 'lrp-' + port_ids[0]):
            # The router name key has been correctly updated; now it
            # contains the router name (starting with "neutron-").
            self.assertEqual(
                lrp['external_ids'][ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY],
                ovn_const.OVN_NAME_PREFIX + router['id'])
            return
        self.fail('Logical_Router_Port lrp-%s not found' % port_ids[0])

    def _get_nb_global_options(self):
        return self.nb_api.db_get(
            'NB_Global', '.', 'options').execute(check_error=True)

    def _test_update_ha_failover(self, strategy, expected_rx, expected_tx,
                                 expected_mult, manual_overrides=None):
        cfg.CONF.set_override('ha_failover_strategy', strategy, group='ovn')
        if manual_overrides:
            for key, value in manual_overrides.items():
                cfg.CONF.set_override(key, value, group='ovn')

        self.assertRaises(periodics.NeverAgain,
                          self.maint.update_ha_failover)

        options = self._get_nb_global_options()
        self.assertEqual(str(expected_rx), options['bfd-min-rx'])
        self.assertEqual(str(expected_tx), options['bfd-min-tx'])
        self.assertEqual(str(expected_mult), options['bfd-mult'])

    def test_update_ha_failover_normal(self):
        self._test_update_ha_failover(
            ovn_const.OVN_HA_FAILOVER_NORMAL,
            ovn_const.OVN_BFD_MIN_RX,
            ovn_const.OVN_BFD_MIN_TX,
            ovn_const.OVN_BFD_MULT)

    def test_update_ha_failover_aggressive(self):
        self._test_update_ha_failover(
            ovn_const.OVN_HA_FAILOVER_AGGRESSIVE,
            ovn_const.OVN_BFD_MIN_RX / 2,
            ovn_const.OVN_BFD_MIN_TX,
            ovn_const.OVN_BFD_MULT)

    def test_update_ha_failover_conservative(self):
        self._test_update_ha_failover(
            ovn_const.OVN_HA_FAILOVER_CONSERVATIVE,
            ovn_const.OVN_BFD_MIN_RX,
            ovn_const.OVN_BFD_MIN_TX,
            ovn_const.OVN_BFD_MULT * 2)

    def test_update_ha_failover_manual(self):
        self._test_update_ha_failover(
            ovn_const.OVN_HA_FAILOVER_MANUAL,
            5000, 400, 8,
            manual_overrides={'bfd_min_rx': 5000,
                              'bfd_min_tx': 400,
                              'bfd_mult': 8})

    def test_update_security_group_with_address_group(self):
        """Test missing Address_Sets are recreated and ACLs are updated.

        Simulates a pre-upgrade state where an Address Group existed in
        Neutron but the corresponding OVN Address_Sets were never created,
        so the ACL match string does not reference the Address_Set.
        """
        # 1. Create an AG, a SG, and two SG rules (IPv4 + IPv6) pointing
        #    to the AG.
        ag = self._create_address_group(
            name='ag_maint_test',
            addresses=['172.16.0.1/32', '2001:db8::1/128'])
        neutron_sg = self._create_security_group()
        sg_rule_v4 = self._create_security_group_rule(
            neutron_sg['id'], remote_address_group_id=ag['id'],
            ethertype=n_const.IPv4)
        sg_rule_v6 = self._create_security_group_rule(
            neutron_sg['id'], remote_address_group_id=ag['id'],
            ethertype=n_const.IPv6)

        # 2. Verify the ACLs reference the Address_Sets and that both
        #    Address_Sets (IPv4, IPv6) exist.
        ag_as_name_v4 = utils.ovn_ag_addrset_name(ag['id'], 'ip4')
        ag_as_name_v6 = utils.ovn_ag_addrset_name(ag['id'], 'ip6')

        acl_v4 = self._find_security_group_rule_row_by_id(sg_rule_v4['id'])
        self.assertIsNotNone(acl_v4)
        acl_v4_uuid_before = acl_v4.uuid
        self.assertIn(ag_as_name_v4, acl_v4.match)

        acl_v6 = self._find_security_group_rule_row_by_id(sg_rule_v6['id'])
        self.assertIsNotNone(acl_v6)
        acl_v6_uuid_before = acl_v6.uuid
        self.assertIn(ag_as_name_v6, acl_v6.match)

        address_sets = self._find_address_set_for_ag(ag['id'])
        self.assertIsNotNone(address_sets[n_const.IP_VERSION_4])
        self.assertIsNotNone(address_sets[n_const.IP_VERSION_6])

        # 3. Simulate the pre-upgrade state: delete the Address_Sets and
        #    strip the Address_Set reference from both ACL matches.
        self._delete_address_set_for_ag(ag['id'])
        address_sets = self._find_address_set_for_ag(ag['id'])
        self.assertIsNone(address_sets[n_const.IP_VERSION_4])
        self.assertIsNone(address_sets[n_const.IP_VERSION_6])

        match_v4_without_ag = acl_v4.match.replace(
            ' && ip4.src == $%s' % ag_as_name_v4, '')
        self.nb_api.db_set(
            'ACL', acl_v4.uuid,
            ('match', match_v4_without_ag)).execute(check_error=True)
        acl_v4 = self._find_security_group_rule_row_by_id(sg_rule_v4['id'])
        self.assertNotIn(ag_as_name_v4, acl_v4.match)

        match_v6_without_ag = acl_v6.match.replace(
            ' && ip6.src == $%s' % ag_as_name_v6, '')
        self.nb_api.db_set(
            'ACL', acl_v6.uuid,
            ('match', match_v6_without_ag)).execute(check_error=True)
        acl_v6 = self._find_security_group_rule_row_by_id(sg_rule_v6['id'])
        self.assertNotIn(ag_as_name_v6, acl_v6.match)

        # 4. Run the maintenance task.
        self.assertRaises(
            periodics.NeverAgain,
            self.maint.update_security_group_with_address_group)

        # 5. Verify the Address_Sets are recreated.
        address_sets = self._find_address_set_for_ag(ag['id'])
        self.assertEqual(['172.16.0.1/32'],
                         address_sets[n_const.IP_VERSION_4].addresses)
        self.assertEqual(['2001:db8::1/128'],
                         address_sets[n_const.IP_VERSION_6].addresses)

        # 6. Verify the previous ACLs are deleted and new ones were created
        #    (different UUIDs) with matches that include the Address_Set
        #    references.
        acl_uuids = set(self.nb_api._tables['ACL'].rows.keys())
        self.assertNotIn(acl_v4_uuid_before, acl_uuids)
        self.assertNotIn(acl_v6_uuid_before, acl_uuids)

        acl_v4_after = self._find_security_group_rule_row_by_id(
            sg_rule_v4['id'])
        self.assertIsNotNone(acl_v4_after)
        self.assertIn(ag_as_name_v4, acl_v4_after.match)

        acl_v6_after = self._find_security_group_rule_row_by_id(
            sg_rule_v6['id'])
        self.assertIsNotNone(acl_v6_after)
        self.assertIn(ag_as_name_v6, acl_v6_after.match)


class TestLogMaintenance(_TestMaintenanceHelper,
                         test_log_driver.LogApiTestCaseBase):
    def test_check_for_logging_conf_change(self):
        # Check no meter exists
        self.assertFalse(self.nb_api._tables['Meter'].rows.values())
        # Add a log object
        self.log_plugin.create_log(self.context, self._log_data())
        # Check a meter and fair meter exist
        self.assertTrue(self.nb_api._tables['Meter'].rows)
        self.assertTrue(self.nb_api._tables['Meter_Band'].rows)
        self.assertEqual(
            len([*self.nb_api._tables['Meter'].rows.values()]),
            len([*self.nb_api._tables['Meter_Band'].rows.values()]))
        self._check_meters_consistency()
        # Update burst and rate limit values on the configuration
        ovn_config.cfg.CONF.set_override('burst_limit', CFG_NEW_BURST,
                                         group='network_log')
        ovn_config.cfg.CONF.set_override('rate_limit', CFG_NEW_RATE,
                                         group='network_log')
        # Call the maintenance task
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_fair_meter_consistency)
        # Check meter band was effectively changed after the maintenance call
        self._check_meters_consistency(CFG_NEW_BURST, CFG_NEW_RATE)

    def _check_meters_consistency(self, new_burst=None, new_rate=None):
        burst, rate = (new_burst, new_rate) if new_burst else (
            cfg.CONF.network_log.burst_limit, cfg.CONF.network_log.rate_limit)
        for meter in [*self.nb_api._tables['Meter'].rows.values()]:
            meter_band = self.nb_api.lookup('Meter_Band', meter.bands[0].uuid)
            if "_stateless" in meter.name:
                self.assertEqual(int(burst / 2), meter_band.burst_size)
                self.assertEqual(int(rate / 2), meter_band.rate)
            else:
                self.assertEqual(burst, meter_band.burst_size)
                self.assertEqual(rate, meter_band.rate)



# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron_lib.ovn import db_sync as db_sync_base
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging
from stevedore import enabled

from neutron.common import config as common_config
from neutron.common.ovn import constants as ovn_const
from neutron.conf.agent import securitygroups_rpc
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.conf.plugins.ml2.drivers.ovn import ovn_db_sync as sync_conf
from neutron import manager
from neutron import opts as neutron_options
from neutron.plugins.ml2.drivers.ovn.mech_driver import mech_driver
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import impl_idl_ovn
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import worker
from neutron.plugins.ml2 import plugin as ml2_plugin

LOG = logging.getLogger(__name__)


class Ml2Plugin(ml2_plugin.Ml2Plugin):

    @property
    def _rpc_workers(self):
        return 0


class OVNMechanismDriver(mech_driver.OVNMechanismDriver):

    def subscribe(self):
        pass

    def post_fork_initialize(self, resource, event, trigger, **kwargs):
        pass

    @property
    def ovn_client(self):
        return self._ovn_client

    def _remove_node_from_hash_ring(self):
        """Don't remove the node from the Hash Ring.

        If this method was not overridden, cleanup would be performed when
        calling the db sync and running neutron server would remove the
        nodes from the Hash Ring.
        """

    # Since we are not using the ovn mechanism driver while syncing,
    # we override the post and pre commit methods so that original ones are
    # not called.
    def create_port_precommit(self, context):
        pass

    def create_port_postcommit(self, context):
        port = context.current
        self.ovn_client.create_port(context.plugin_context, port)

    def update_port_precommit(self, context):
        pass

    def update_port_postcommit(self, context):
        port = context.current
        original_port = context.original
        self.ovn_client.update_port(context.plugin_context, port,
                                    original_port)

    def delete_port_precommit(self, context):
        pass

    def delete_port_postcommit(self, context):
        port = copy.deepcopy(context.current)
        port['network'] = context.network.current
        self.ovn_client.delete_port(context.plugin_context, port['id'])


def setup_conf():
    conf = cfg.CONF
    common_config.register_common_config_options()
    ovn_conf.register_opts()
    ml2_group, ml2_opts = neutron_options.list_ml2_conf_opts()[0]
    cfg.CONF.register_cli_opts(ml2_opts, ml2_group)
    cfg.CONF.register_cli_opts(securitygroups_rpc.security_group_opts,
                               'SECURITYGROUP')
    ovn_group, ovn_opts = ovn_conf.list_opts()[0]
    cfg.CONF.register_cli_opts(ovn_opts, group=ovn_group)
    db_group, neutron_db_opts = db_options.list_opts()[0]
    cfg.CONF.register_cli_opts(neutron_db_opts, db_group)
    sync_conf.register_ovn_db_sync_cli_opts(cfg.CONF)
    # Override Nova notify configuration LP: #1882020
    cfg.CONF.set_override('notify_nova_on_port_status_changes', False)
    cfg.CONF.set_override('notify_nova_on_port_data_changes', False)
    return conf


def _load_drivers(entry_point, driver_name=None):

    def load_driver(ext):
        if not issubclass(ext.plugin, db_sync_base.BaseOvnDbSynchronizer):
            LOG.error("Extension '%s' is not an instance of "
                      "%s and will not be loaded",
                      ext.name, db_sync_base.BaseOvnDbSynchronizer)
            return False
        if driver_name is None:
            return True
        return ext.name == driver_name

    return enabled.EnabledExtensionManager(
        entry_point,
        check_func=load_driver,
        invoke_on_load=False)


def load_synchronize_drivers(driver_name=None):
    return _load_drivers('neutron.ovn.db_sync', driver_name)


def load_db_migration_drivers(driver_name=None):
    return _load_drivers('neutron.ovn.db_migration', driver_name)


def prepare_additional_configuration(conf, mgr):
    for ext in mgr:
        # TODO(slaweq): Remove that check when
        # patch https://review.opendev.org/c/openstack/neutron-lib/+/980435
        # will be available in the neutron-lib and
        # 'prepare_additional_configuration' method will be available in the
        # BaseOvnDbSynchronizer
        if hasattr(ext.plugin, "prepare_additional_configuration"):
            ext.plugin.prepare_additional_configuration(conf)


def configure_mechanism_drivers(conf, mgr):
    required_mechanism_drivers = set()
    for ext in mgr:
        required_mechanism_drivers |= set(
            ext.plugin.get_required_mechanism_drivers())

    conf.set_override(
        'mechanism_drivers', list(required_mechanism_drivers), 'ml2')


def configure_service_plugins(conf, mgr):
    required_plugins = set()
    for ext in mgr:
        required_plugins |= set(ext.plugin.get_required_service_plugins())

    conf.set_override('service_plugins', list(required_plugins))


def configure_ml2_extension_drivers(conf, mgr):
    required_extension_drivers = set()
    for ext in mgr:
        required_extension_drivers |= set(
            ext.plugin.get_required_ml2_extension_drivers())

    extension_drivers = list(
        set(conf.ml2.extension_drivers) | required_extension_drivers)
    conf.set_override('extension_drivers', extension_drivers, 'ml2')


def synchronize_ovn_dbs(mgr, core_plugin, ovn_driver, mode):
    LOG.info('Neutron OVN DBs sync started with mode: %s', mode)
    for sync_driver in mgr:
        LOG.info('Starting synchronize with %s driver',
                 sync_driver.name)
        sync_obj = sync_driver.plugin(core_plugin, ovn_driver, mode)
        sync_obj.do_sync()
        LOG.info('Driver %s sync completed', sync_driver.name)
    LOG.info('Neutron OVN DBs sync completed')


def migrate_neutron_dbs_to_ovn(drv):
    LOG.info("Database migration from OVS to OVN by plugin %s started",
             drv.name)
    # This will call function defined in the ovn.db_migration entry_point,
    # for example for neutron it is migrate_database_to_ovn() function
    drv.plugin()
    LOG.info("Database migration from OVS to OVN by plugin %s completed",
             drv.name)

def main():
    """Main method for syncing neutron networks and ports with ovn nb db.

    This script provides a utility for syncing the OVN Northbound Database
    with the Neutron database.

    This script is used for the migration from ML2/OVS to ML2/OVN.
    """
    conf = setup_conf()

    # if no config file is passed or no configuration options are passed
    # then load configuration from /etc/neutron/neutron.conf
    try:
        conf(project='neutron')
    except TypeError:
        LOG.error('Error parsing the configuration values. Please verify.')
        raise SystemExit(1)

    logging.setup(conf, 'neutron_ovn_db_sync_util', fix_eventlet=False)
    LOG.info('Neutron OVN DB sync started')
    sync_ext_mgr = load_synchronize_drivers(conf.sync_plugin)
    if not sync_ext_mgr.names():
        LOG.error('No OVN DB sync plugin found')
        raise SystemExit(1)

    LOG.info('Loaded sync plugins: %s', ', '.join(sync_ext_mgr.names()))

    mode = ovn_conf.get_ovn_neutron_sync_mode()
    # Migrate mode will run as repair mode in the synchronizer
    migrate = False
    if mode == ovn_const.OVN_DB_SYNC_MODE_MIGRATE:
        mode = ovn_const.OVN_DB_SYNC_MODE_REPAIR
        migrate = True
    if mode not in [ovn_const.OVN_DB_SYNC_MODE_LOG,
                    ovn_const.OVN_DB_SYNC_MODE_REPAIR]:
        LOG.error(
            'Invalid sync mode: ["%s"]. Should be "%s" or "%s"',
            mode,
            ovn_const.OVN_DB_SYNC_MODE_LOG,
            ovn_const.OVN_DB_SYNC_MODE_REPAIR)
        raise SystemExit(1)

    # Validate and modify core plugin and ML2 mechanism drivers for syncing.
    if (conf.core_plugin.endswith('.Ml2Plugin') or
            conf.core_plugin == 'ml2'):
        conf.core_plugin = (
            'neutron.cmd.ovn.neutron_ovn_db_sync_util.Ml2Plugin')
        if not conf.ml2.mechanism_drivers:
            LOG.error('Please use --config-file to specify '
                      'neutron and ml2 configuration file.')
            raise SystemExit(1)
        if 'ovn' not in conf.ml2.mechanism_drivers:
            LOG.error('No "ovn" mechanism driver found: "%s".',
                      conf.ml2.mechanism_drivers)
            raise SystemExit(1)
        configure_mechanism_drivers(conf, sync_ext_mgr)
        configure_service_plugins(conf, sync_ext_mgr)
        configure_ml2_extension_drivers(conf, sync_ext_mgr)
        prepare_additional_configuration(conf, sync_ext_mgr)

    else:
        LOG.error('Invalid core plugin: ["%s"].', conf.core_plugin)
        raise SystemExit(1)

    mech_worker = worker.MaintenanceWorker
    try:
        ovn_nb_api = impl_idl_ovn.OvsdbNbOvnIdl.from_worker(mech_worker)
    except RuntimeError:
        LOG.error('Invalid --ovn-ovn_nb_connection parameter provided.')
        raise SystemExit(1)

    try:
        ovn_sb_api = impl_idl_ovn.OvsdbSbOvnIdl.from_worker(mech_worker)
    except RuntimeError:
        LOG.error('Invalid --ovn-ovn_sb_connection parameter provided.')
        raise SystemExit(1)

    manager.init()
    core_plugin = directory.get_plugin()
    driver = core_plugin.mechanism_manager.mech_drivers['ovn-sync']
    # The L3 code looks for the OVSDB connection on the 'ovn' driver
    # and will fail with a KeyError if it isn't there
    core_plugin.mechanism_manager.mech_drivers['ovn'] = driver
    ovn_driver = driver.obj
    ovn_driver.nb_ovn = ovn_nb_api
    ovn_driver.sb_ovn = ovn_sb_api
    ovn_driver._post_fork_event.set()

    synchronize_ovn_dbs(
        sync_ext_mgr, core_plugin, ovn_driver, mode
    )

    # TODO(slaweq): add drivers for the ovs2ovn migration
    if migrate:
        LOG.info("Neutron database migration from OVS to OVN started")
        migration_drivers = load_db_migration_drivers(conf.migration_plugin)
        migration_drivers.map(migrate_neutron_dbs_to_ovn)
        LOG.info("Neutron database migration from OVS to OVN completed")

    LOG.info('Neutron OVN DB sync completed')



# Copyright (c) 2023 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import re

import netaddr
from neutron_lib import constants
from neutron_lib.plugins.ml2 import ovs_constants as p_const
from neutron_lib.plugins import utils as p_utils
from oslo_log import log as logging
from oslo_utils import uuidutils

from neutron.tests.common.exclusive_resources import ip_network
from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.fullstack.resources import machine

LOG = logging.getLogger(__name__)

METADATA_REQUEST_TIMEOUT = 60
METADATA_REQUEST_SLEEP = 5
TOO_MANY_REQUESTS_CODE = '429'


class OvsMetadataExtensionTestCase(base.BaseFullStackTestCase):
    number_of_hosts = 1

    def setUp(self):
        host_desc = [
            environment.HostDescription(
                l2_agent_type=constants.AGENT_TYPE_OVS,
                firewall_driver='openvswitch',
                l3_agent=True,
                dhcp_agent=False) for _ in range(self.number_of_hosts)]
        env_desc = environment.EnvironmentDescription(
            mech_drivers='openvswitch',
            has_metadata=True, metadata_host='127.0.0.1',
            metadata_port=58775,
            host_proxy_listen_port=55555,
            enable_traditional_dhcp=False)
        env = environment.Environment(env_desc, host_desc)
        super().setUp(env)
        self.tenant_id = uuidutils.generate_uuid()
        self.vm_id_1 = uuidutils.generate_uuid()
        self.vm_id_2 = uuidutils.generate_uuid()

        network = self.safe_client.create_network(
            self.tenant_id, name='public', external=True)
        cidr = self.useFixture(
            ip_network.ExclusiveIPNetwork(
                "240.0.0.0", "240.255.255.255", "24")).network
        self.safe_client.create_subnet(
            self.tenant_id, network['id'], cidr)

        router = self.safe_client.create_router(
            self.tenant_id, external_network=network['id'])

        self.network = self.safe_client.create_network(
            self.tenant_id, 'network-test')
        subnet_routes_v4 = [
            {"destination": "1.1.1.0/24", "nexthop": "10.0.0.100"},
            {"destination": "2.2.2.2/32", "nexthop": "10.0.0.101"}]
        self.subnet_v4 = self.safe_client.create_subnet(
            self.tenant_id, self.network['id'],
            cidr='10.0.0.0/24',
            gateway_ip='10.0.0.1',
            enable_dhcp=False,
            name='subnet-v4-test',
            host_routes=subnet_routes_v4)

        router_interface_info = self.safe_client.add_router_interface(
            router['id'], self.subnet_v4['id'])
        self.block_until_port_status_active(
            router_interface_info['port_id'])

        subnet_routes_v6 = [
            {"destination": "2001:4860:4860::8888/128",
             "nexthop": "fda7:a5cc:3460:1::1"},
            {"destination": "1234:5678:abcd::/64",
             "nexthop": "fda7:a5cc:3460:1::fff"}]
        self.subnet_v6 = self.safe_client.create_subnet(
            self.tenant_id, self.network['id'],
            cidr='fda7:a5cc:3460:1::/64',
            gateway_ip='fda7:a5cc:3460:1::1',
            enable_dhcp=True,
            ipv6_address_mode="dhcpv6-stateful",
            ipv6_ra_mode="dhcpv6-stateful",
            ip_version=6,
            name='subnet-v6-test',
            host_routes=subnet_routes_v6)

        # Need router radvd to send IPv6 address prefix to make the default
        # route work.
        router_interface_info = self.safe_client.add_router_interface(
            router['id'], self.subnet_v6['id'])
        self.block_until_port_status_active(
            router_interface_info['port_id'])

    def block_until_port_status_active(self, port_id):
        def is_port_status_active():
            port = self.client.show_port(port_id)
            return port['port']['status'] == 'ACTIVE'
        base.wait_until_true(lambda: is_port_status_active(), sleep=1)

    def _prepare_vms(self):
        sgs = [self.safe_client.create_security_group(self.tenant_id)
               for _ in range(2)]

        port1 = self.safe_client.create_port(
            self.tenant_id, self.network['id'],
            self.environment.hosts[0].hostname,
            device_owner="compute:test_ovs_meta_1",
            device_id=self.vm_id_1,
            security_groups=[sgs[0]['id']])

        port2 = self.safe_client.create_port(
            self.tenant_id, self.network['id'],
            self.environment.hosts[0].hostname,
            device_owner="compute:test_ovs_meta_2",
            device_id=self.vm_id_2,
            security_groups=[sgs[1]['id']])

        # insert security-group-rules allow icmp
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[0]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv4,
            protocol=constants.PROTO_NAME_ICMP)
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[0]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv6,
            protocol=constants.PROTO_NAME_ICMP)

        # insert security-group-rules allow icmp
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[1]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv4,
            protocol=constants.PROTO_NAME_ICMP)
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[1]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv6,
            protocol=constants.PROTO_NAME_ICMP)

        vm1 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client,
                neutron_port=port1,
                use_dhcp=False,
                use_dhcp6=False))

        vm2 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client,
                neutron_port=port2,
                use_dhcp=False,
                use_dhcp6=False))
        return machine.FakeFullstackMachinesList([vm1, vm2])

    def _wait_for_metadata_flows_applied(self, vm, table, actions):

        def _is_metadata_flow_set(vm, table, actions):
            LOG.info("Metadata bridge verify actions: %s", actions)
            flows = vm.host.br_meta.dump_flows_for_table(table)
            flows_list = flows.splitlines()
            LOG.info("Metadata bridge flows_list: %s", flows_list)
            pattern = re.compile(
                r"^.* table={},.* actions={}".format(table,
                                                     re.escape(actions)))
            for flow in flows_list:
                if pattern.match(flow.strip()):
                    return True
            return False
        base.wait_until_true(lambda: _is_metadata_flow_set(
            vm, table, actions))

    def test_ovs_meta_agent_extension_verify_ovs_flows(self):
        vms = self._prepare_vms()
        vms.block_until_all_boot()

        # Check ovs flows
        vm_0_provider_ip = vms[0].bridge.get_value_from_other_config(
            vms[0].port.name, "provider_ip", value_type=str)
        vm_0_provider_mac = vms[0].bridge.get_value_from_other_config(
            vms[0].port.name, "provider_mac", value_type=str)

        actions_0 = ("strip_vlan,mod_dl_src:%s,"
                     "mod_nw_src:%s,resubmit(,87)") % (vm_0_provider_mac,
                                                       vm_0_provider_ip)
        self._wait_for_metadata_flows_applied(vms[0], 80, actions_0)

        vm_1_provider_ip = vms[1].bridge.get_value_from_other_config(
            vms[1].port.name, "provider_ip", value_type=str)
        vm_1_provider_mac = vms[1].bridge.get_value_from_other_config(
            vms[1].port.name, "provider_mac", value_type=str)
        actions_1 = ("strip_vlan,mod_dl_src:%s,"
                     "mod_nw_src:%s,resubmit(,87)") % (vm_1_provider_mac,
                                                       vm_1_provider_ip)
        self._wait_for_metadata_flows_applied(vms[1], 80, actions_1)

        tap_meta_ofport = vms[0].host.br_meta.get_port_ofport("tap-meta")
        self._wait_for_metadata_flows_applied(
            vms[0], 87,
            ("mod_dl_dst:fa:16:ee:00:00:01,mod_nw_dst:240.0.0.1,"
             "mod_tp_dst:55555,output:%s" % tap_meta_ofport))

        vm_0_provider_mac_hex = "0x%s" % (
            vm_0_provider_mac.replace(":", ""))
        vm_0_provider_ip_hex = "%x" % netaddr.IPAddress(vm_0_provider_ip)
        vm_0_arp_res_actions = (
            "load:0x2->NXM_OF_ARP_OP[],"
            "move:NXM_NX_ARP_SHA[]->NXM_NX_ARP_THA[],"
            "move:NXM_OF_ARP_SPA[]->NXM_OF_ARP_TPA[],"
            "load:%s->NXM_NX_ARP_SHA[],"
            "load:0x%s->NXM_OF_ARP_SPA[],"
            "move:NXM_OF_ETH_SRC[]->NXM_OF_ETH_DST[],"
            "mod_dl_src:%s,IN_PORT") % (vm_0_provider_mac_hex,
                                        vm_0_provider_ip_hex,
                                        vm_0_provider_mac)
        self._wait_for_metadata_flows_applied(
            vms[0], 90, vm_0_arp_res_actions)

        vm_1_provider_mac_hex = "0x%s" % (vm_1_provider_mac.replace(":", ""))
        vm_1_provider_ip_hex = "%x" % netaddr.IPAddress(vm_1_provider_ip)
        vm_1_arp_res_actions = (
            "load:0x2->NXM_OF_ARP_OP[],"
            "move:NXM_NX_ARP_SHA[]->NXM_NX_ARP_THA[],"
            "move:NXM_OF_ARP_SPA[]->NXM_OF_ARP_TPA[],"
            "load:%s->NXM_NX_ARP_SHA[],"
            "load:0x%s->NXM_OF_ARP_SPA[],"
            "move:NXM_OF_ETH_SRC[]->NXM_OF_ETH_DST[],"
            "mod_dl_src:%s,IN_PORT") % (vm_1_provider_mac_hex,
                                        vm_1_provider_ip_hex,
                                        vm_1_provider_mac)
        self._wait_for_metadata_flows_applied(
            vms[1], 90, vm_1_arp_res_actions)

        local_vlan = vms[0].bridge.get_port_tag_by_name(vms[0].port.name)

        patch_name = p_utils.get_interface_name(
            vms[0].host.br_meta.br_name,
            prefix=p_const.PEER_PHYSICAL_PREFIX)
        patch_ofport = vms[0].host.br_meta.get_port_ofport(patch_name)
        self._wait_for_metadata_flows_applied(
            vms[0], 91,
            ("mod_vlan_vid:%s,mod_dl_dst:%s,"
             "mod_nw_src:169.254.169.254,mod_nw_dst:%s,mod_tp_src:80,"
             "output:%s" % (local_vlan, vms[0].neutron_port['mac_address'],
                            vms[0].ip, patch_ofport)))
        local_vlan = vms[1].bridge.get_port_tag_by_name(vms[1].port.name)
        patch_name = p_utils.get_interface_name(
            vms[1].host.br_meta.br_name,
            prefix=p_const.PEER_PHYSICAL_PREFIX)
        patch_ofport = vms[1].host.br_meta.get_port_ofport(patch_name)
        self._wait_for_metadata_flows_applied(
            vms[1], 91,
            ("mod_vlan_vid:%s,mod_dl_dst:%s,"
             "mod_nw_src:169.254.169.254,mod_nw_dst:%s,mod_tp_src:80,"
             "output:%s" % (local_vlan, vms[1].neutron_port['mac_address'],
                            vms[1].ip, patch_ofport)))



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from oslo_config import cfg
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import idlutils


from neutron.agent.linux import ip_lib
from neutron.common.ovn import constants as ovn_const
from neutron.conf.services import bgp as bgp_config
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import (
    commands as ovn_commands)
from neutron.services.bgp import commands
from neutron.services.bgp import constants
from neutron.services.bgp import exceptions
from neutron.services.bgp import helpers
from neutron.tests.functional.services import bgp


def _get_unique_name(prefix="test"):
    return f"{prefix}_{uuidutils.generate_uuid()[:8]}"


def _create_fake_chassis():
    return _FakeChassis(_get_unique_name("chassis"))


class _FakeChassis:
    def __init__(self, name):
        self.uuid = uuidutils.generate_uuid()
        self.name = name


class _AddBaseCommand:
    table = None
    command = None

    def _assert_table_row_exists(self, name, should_exist=True):
        try:
            row = self.nb_api.lookup(self.table, name)
            if should_exist:
                self.assertIsNotNone(row)
                self.assertEqual(row.name, name)
                return row
            self.fail(f"{self.table} {name} should not exist")
        except idlutils.RowNotFound:
            if should_exist:
                self.fail(f"{self.table} {name} not found")
            return None

    def create_row(self, name, **kwargs):
        self.command(self.nb_api, name, **kwargs).execute(check_error=True)

    def test_create_new_row(self):
        name = _get_unique_name()

        # Verify row doesn't exist initially
        self._assert_table_row_exists(name, should_exist=False)

        self.create_row(name)

        # Verify the row was created
        created_row = self._assert_table_row_exists(name)
        self.assertEqual(created_row.name, name)

    def test_create_existing_row(self):
        name = _get_unique_name()

        # Create row first time
        self.create_row(name)

        # Verify first creation worked
        row1 = self._assert_table_row_exists(name)

        # Create same row again - should be idempotent
        self.create_row(name)

        # Lookup should not fail with duplicated name
        row2 = self._assert_table_row_exists(name)
        self.assertEqual(row1.uuid, row2.uuid)

    def test_external_ids_update_on_create(self):
        name = _get_unique_name()
        self.create_row(name, external_ids={'id1': 'value1'})
        row = self._assert_table_row_exists(name)
        self.assertEqual(row.external_ids.get('id1'), 'value1')

        # Create same row again - should be idempotent
        self.create_row(name, external_ids={'id1': 'value2'})
        row = self._assert_table_row_exists(name)
        self.assertEqual(row.external_ids.get('id1'), 'value2')

        # Create with different name - should create a new row
        name = _get_unique_name()
        self.create_row(name, external_ids={'id1': 'value3'})
        row = self._assert_table_row_exists(name)


class TestGetGwIps(bgp.BaseBgpNbIdlTestCase):
    def _add_dhcp_options_command(self, net_id, cidr, router_ip=None):
        options = {}
        if router_ip:
            options['router'] = router_ip
        return ovn_commands.AddDHCPOptionsCommand(
            self.nb_api, net_id, may_exist=False,
            cidr=cidr,
            options=options,
            external_ids={ovn_const.OVN_NETWORK_ID_EXT_ID_KEY: net_id},
        )

    def test_get_gw_ips_for_switch_returns_empty_when_no_dhcp_options(self):
        net_id = uuidutils.generate_uuid()
        ls_name = f"neutron-{net_id}"
        ls = self.nb_api.ls_add(ls_name).execute(check_error=True)

        gw_ips = commands._get_gw_ips_for_switch(self.nb_api, ls)

        self.assertEqual([], gw_ips)

    def test_get_gw_ips_for_switch_returns_gateway_ip_from_dhcp_options(self):
        net_id = uuidutils.generate_uuid()
        ls_name = f"neutron-{net_id}"

        with self.nb_api.transaction(check_error=True) as txn:
            ls = txn.add(self.nb_api.ls_add(ls_name))
            txn.add(self._add_dhcp_options_command(
                net_id, '10.0.0.0/24', '10.0.0.5'))

        gw_ips = commands._get_gw_ips_for_switch(self.nb_api, ls.result)

        self.assertEqual(['10.0.0.5/24'], gw_ips)

    def test_get_gw_ips_for_switch_ignores_other_networks(self):
        net_id = uuidutils.generate_uuid()
        other_net_id = uuidutils.generate_uuid()
        ls_name = f"neutron-{net_id}"

        with self.nb_api.transaction(check_error=True) as txn:
            ls = txn.add(self.nb_api.ls_add(ls_name))
            txn.add(self._add_dhcp_options_command(
                other_net_id, '10.0.0.0/24', '10.0.0.1'))

        gw_ips = commands._get_gw_ips_for_switch(self.nb_api, ls.result)

        self.assertEqual([], gw_ips)

    def test_get_gw_ips_for_switch_returns_multiple_gateways(self):
        net_id = uuidutils.generate_uuid()
        ls_name = f"neutron-{net_id}"

        with self.nb_api.transaction(check_error=True) as txn:
            ls = txn.add(self.nb_api.ls_add(ls_name))
            txn.add(self._add_dhcp_options_command(
                net_id, '10.0.0.0/24', '10.0.0.1'))
            txn.add(self._add_dhcp_options_command(
                net_id, '2001:db8::/64', '2001:db8::1'))

        gw_ips = commands._get_gw_ips_for_switch(self.nb_api, ls.result)

        self.assertCountEqual(['10.0.0.1/24', '2001:db8::1/64'], gw_ips)

    def test_get_gw_ips_for_switch_skips_dhcp_options_without_router(self):
        net_id = uuidutils.generate_uuid()
        ls_name = f"neutron-{net_id}"

        with self.nb_api.transaction(check_error=True) as txn:
            ls = txn.add(self.nb_api.ls_add(ls_name))
            txn.add(self._add_dhcp_options_command(
                net_id, '10.0.0.0/24'))

        gw_ips = commands._get_gw_ips_for_switch(self.nb_api, ls.result)

        self.assertEqual([], gw_ips)


class TestCaseWithOutputPort(bgp.BaseBgpNbIdlTestCase):
    def _is_bgp_supported(self):
        return bgp.is_policy_output_port_column_supported(self.nb_api.idl)


class LsAddCommandTestCase(bgp.BaseBgpNbIdlTestCase, _AddBaseCommand):
    table = 'Logical_Switch'
    command = commands._LsAddCommand


class LrAddCommandTestCase(bgp.BaseBgpNbIdlTestCase,
                           _AddBaseCommand):
    table = 'Logical_Router'
    command = commands._LrAddCommand


class LspAddCommandTestCase(bgp.BaseBgpNbIdlTestCase, _AddBaseCommand):
    table = 'Logical_Switch_Port'
    command = commands._LspAddCommand

    def setUp(self):
        super().setUp()
        self.ls_name = _get_unique_name()
        self.nb_api.ls_add(self.ls_name).execute(check_error=True)

    def create_row(self, name, **kwargs):
        return self.command(self.nb_api, self.ls_name, name, **kwargs).execute(
            check_error=True)

    def test_create_existing_with_different_attributes(self):
        name = _get_unique_name()
        self.create_row(
            name, options={'peer-port': 'lsp-peer-1'},
            external_ids={'id1': 'value1'})
        lsp = self._assert_table_row_exists(name)
        self.assertEqual(lsp.options.get('peer-port'), 'lsp-peer-1')
        self.assertEqual(lsp.external_ids.get('id1'), 'value1')

        # Should update the options
        self.create_row(name, options={'peer-port': 'lsp-peer-2'},
                        external_ids={'id1': 'value2'})
        lsp = self._assert_table_row_exists(name)
        self.assertEqual(lsp.options.get('peer-port'), 'lsp-peer-2')
        self.assertEqual(lsp.external_ids.get('id1'), 'value2')


class LrpAddCommandTestCase(bgp.BaseBgpNbIdlTestCase, _AddBaseCommand):
    table = 'Logical_Router_Port'
    command = commands._LrpAddCommand

    def setUp(self):
        super().setUp()
        self.lr_name = _get_unique_name()
        self.nb_api.lr_add(self.lr_name).execute(check_error=True)

    def create_row(self, name, **kwargs):
        return self.command(
            self.nb_api, self.lr_name, name, **kwargs).execute(
                check_error=True)

    def test_create_existing_with_different_attributes(self):
        name = _get_unique_name()
        self.create_row(name, networks=['192.168.1.0/24'], peer='lrp-peer-1')
        lrp = self._assert_table_row_exists(name)

        bad_mac = '00:00:00:00:00:00'

        # Set a different MAC address
        self.nb_api.db_set(
            'Logical_Router_Port', name,
            mac=bad_mac).execute(check_error=True)
        self.assertEqual(lrp.mac, bad_mac)
        self.assertEqual(lrp.networks, ['192.168.1.0/24'])
        self.assertEqual(lrp.peer, ['lrp-peer-1'])

        # Should update the MAC address
        self.create_row(name, networks=['192.168.2.0/24'], peer='lrp-peer-2')
        lrp = self._assert_table_row_exists(name)
        self.assertNotEqual(lrp.mac, bad_mac)
        self.assertEqual(lrp.networks, ['192.168.2.0/24'])
        self.assertEqual(lrp.peer, ['lrp-peer-2'])


class HAChassisGroupAddCommandTestCase(bgp.BaseBgpNbIdlTestCase,
                                       _AddBaseCommand):
    table = 'HA_Chassis_Group'
    command = commands._HAChassisGroupAddCommand


class CreateSwitchWithLocalnetCommandTestCase(bgp.BaseBgpNbIdlTestCase):
    def _validate_localnet_port(self, ls_name, network_name):
        """Validate localnet port was created correctly"""
        localnet_lsp_name = helpers.get_lsp_localnet_name(ls_name)
        lsp = self.nb_api.lookup('Logical_Switch_Port', localnet_lsp_name)
        self.assertEqual(lsp.type, 'localnet')
        self.assertEqual(lsp.options.get('network_name'), network_name)
        self.assertEqual(lsp.addresses, ['unknown'])

    def _create_lsp(self, ls_name, lsp_name, **attrs):
        """Helper to create LSP with wrong attributes"""
        self.nb_api.lsp_add(ls_name, lsp_name, **attrs).execute(
            check_error=True)

    def test_create_new_switch_with_localnet(self):
        ls_name = _get_unique_name()
        network_name = 'test-network'

        commands.CreateSwitchWithLocalnetCommand(
            self.nb_api, ls_name, network_name).execute(check_error=True)

        ls = self.nb_api.ls_get(ls_name).execute(check_error=True)
        self.assertEqual(ls.name, ls_name)

        self._validate_localnet_port(ls_name, network_name)

    def test_create_existing_switch_updates_localnet(self):
        ls_name = _get_unique_name()
        network_name = 'test-network'

        # Create switch first
        self.nb_api.ls_add(ls_name).execute(check_error=True)

        # Execute command
        commands.CreateSwitchWithLocalnetCommand(
            self.nb_api, ls_name, network_name).execute(check_error=True)

        # Verify localnet port was created even with existing switch
        self._validate_localnet_port(ls_name, network_name)

    def test_create_with_existing_localnet_wrong_attributes(self):
        """Test corner case where localnet port exists with wrong attributes"""
        ls_name = _get_unique_name()
        network_name = 'test-network'

        # Create switch and localnet port with wrong attributes
        self.nb_api.ls_add(ls_name).execute(check_error=True)
        localnet_lsp_name = helpers.get_lsp_localnet_name(ls_name)
        self._create_lsp(
            ls_name, localnet_lsp_name,
            type='patch',  # wrong type
            options={'wrong': 'value'},  # wrong options
            addresses=['00:00:00:00:00:01']  # wrong addresses
        )

        # Execute command should fix the attributes
        commands.CreateSwitchWithLocalnetCommand(
            self.nb_api, ls_name, network_name).execute(check_error=True)

        # Verify attributes were corrected
        self._validate_localnet_port(ls_name, network_name)


class CreateLspLocalnetCommandTestCase(bgp.BaseBgpNbIdlTestCase):
    def _validate_localnet_port(self, ls_name, network_name):
        """Validate localnet port was created correctly"""
        localnet_lsp_name = helpers.get_lsp_localnet_name(ls_name)
        lsp = self.nb_api.lookup('Logical_Switch_Port', localnet_lsp_name)
        self.assertEqual(lsp.type, 'localnet')
        self.assertEqual(lsp.options.get('network_name'), network_name)
        self.assertEqual(lsp.addresses, ['unknown'])
        return lsp

    def _create_lsp(self, ls_name, lsp_name, **attrs):
        self.nb_api.lsp_add(ls_name, lsp_name, **attrs).execute(
            check_error=True)

    def setUp(self):
        super().setUp()
        self.ls_name = _get_unique_name()
        self.nb_api.ls_add(self.ls_name).execute(check_error=True)

    def test_create_localnet_port(self):
        network_name = 'test-network'

        commands.CreateLspLocalnetCommand(
            self.nb_api, self.ls_name, network_name).execute(check_error=True)

        self._validate_localnet_port(self.ls_name, network_name)

    def test_update_existing_localnet_with_different_network(self):
        network_name1 = 'test-network-1'
        network_name2 = 'test-network-2'

        # Create first localnet port
        commands.CreateLspLocalnetCommand(
            self.nb_api, self.ls_name, network_name1).execute(check_error=True)

        # Update with different network name
        commands.CreateLspLocalnetCommand(
            self.nb_api, self.ls_name, network_name2).execute(check_error=True)

        # Verify network name was updated
        self._validate_localnet_port(self.ls_name, network_name2)

    def test_fix_localnet_with_wrong_type_and_options(self):
        network_name = 'test-network'
        localnet_lsp_name = helpers.get_lsp_localnet_name(self.ls_name)

        self._create_lsp(
            self.ls_name, localnet_lsp_name,
            type='router',  # wrong type
            options={'router-port': 'wrong'},  # wrong options
            addresses=['router']  # wrong addresses
        )

        commands.CreateLspLocalnetCommand(
            self.nb_api, self.ls_name, network_name).execute(check_error=True)

        lsp = self._validate_localnet_port(self.ls_name, network_name)
        self.assertNotIn('router-port', lsp.options)


class LrPolicyAddCommandTestCase(TestCaseWithOutputPort):
    table = 'Logical_Router_Policy'
    command = commands._LrPolicyAddCommand

    def setUp(self):
        super().setUp()
        self.router_name = _get_unique_name()
        self.lrp_inport_name = _get_unique_name()
        self.lrp_nexthop_name = _get_unique_name()
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.lr_add(self.router_name))
            txn.add(self.nb_api.lrp_add(
                self.router_name,
                self.lrp_inport_name,
                mac="00:00:00:00:00:01",
                networks=["192.168.1.1/30"]))
            txn.add(self.nb_api.lrp_add(
                self.router_name,
                self.lrp_nexthop_name,
                mac="00:00:00:00:00:02",
                networks=["192.168.1.2/30"]))

    @bgp.requires_ovn_version_with_bgp()
    def test_create_new_policy(self):
        priority = 100
        match = 'inport==\"%s\"' % self.lrp_inport_name
        action = 'reroute'
        nexthop_port = self.nb_api.lrp_get(self.lrp_nexthop_name).execute(
            check_error=True)
        commands._LrPolicyAddCommand(
            self.nb_api,
            router=self.router_name,
            priority=priority,
            match=match,
            action=action,
            output_port=nexthop_port).execute(check_error=True)

        expected_exthops = [ip_lib.get_ipv6_lladdr(nexthop_port.mac)]
        policy = self.nb_api.lr_policy_list(
            self.router_name).execute(check_error=True)[0]
        self.assertEqual(priority, policy.priority)
        self.assertEqual(self.lrp_nexthop_name, policy.output_port[0].name)
        self.assertCountEqual(expected_exthops, policy.nexthops)

    @bgp.requires_ovn_version_with_bgp()
    def test_create_existing_policy(self):
        priority = 100
        match = 'inport==\"%s\"' % self.lrp_inport_name
        action = 'reroute'
        nexthop_port = self.nb_api.lrp_get(self.lrp_nexthop_name).execute(
            check_error=True)

        commands._LrPolicyAddCommand(
            self.nb_api,
            router=self.router_name,
            priority=priority,
            match=match,
            action=action,
            output_port=nexthop_port,
            nexthops=['192.168.1.2']).execute(check_error=True)

        commands._LrPolicyAddCommand(
            self.nb_api,
            router=self.router_name,
            priority=priority,
            match=match,
            action=action,
            output_port=nexthop_port).execute(check_error=True)

        expected_exthops = [ip_lib.get_ipv6_lladdr(nexthop_port.mac)]
        policy = self.nb_api.lr_policy_list(
            self.router_name).execute(check_error=True)[0]
        self.assertEqual(priority, policy.priority)
        self.assertEqual(self.lrp_nexthop_name, policy.output_port[0].name)
        self.assertCountEqual(expected_exthops, policy.nexthops)

        commands._LrPolicyAddCommand(
            self.nb_api,
            router=self.router_name,
            priority=priority,
            match=match,
            action=action,
            output_port=nexthop_port).execute(check_error=True)

        policies = self.nb_api.lr_policy_list(
            self.router_name).execute(check_error=True)
        self.assertEqual(1, len(policies))

    @bgp.requires_ovn_version_with_bgp()
    def test_create_policy_with_nexthops(self):
        priority = 100
        match = 'inport==\"%s\"' % self.lrp_inport_name
        action = 'reroute'
        output_port = self.nb_api.lrp_get(self.lrp_nexthop_name).execute(
            check_error=True)
        nexthops = ['192.168.1.2']
        commands._LrPolicyAddCommand(
            self.nb_api,
            router=self.router_name,
            priority=priority,
            match=match,
            action=action,
            output_port=output_port,
            nexthops=nexthops).execute(check_error=True)

        policy = self.nb_api.lr_policy_list(
            self.router_name).execute(check_error=True)[0]
        self.assertEqual(priority, policy.priority)
        self.assertEqual(match, policy.match)
        self.assertEqual(action, policy.action)
        self.assertEqual(self.lrp_nexthop_name, policy.output_port[0].name)
        self.assertCountEqual(nexthops, policy.nexthops)


class ReconcileRouterCommandTestCase(bgp.BaseBgpNbIdlTestCase):
    def _validate_router_created(self, router_name):
        router = self.nb_api.lr_get(router_name).execute(check_error=True)
        self.assertEqual(router.name, router_name)

    def test_reconcile_new_router(self):
        router_name = _get_unique_name()

        commands.ReconcileRouterCommand(
            self.nb_api, router_name).execute(check_error=True)

        self._validate_router_created(router_name)

    def test_reconcile_existing_router(self):
        router_name = _get_unique_name()

        self.nb_api.lr_add(router_name).execute(check_error=True)

        commands.ReconcileRouterCommand(
            self.nb_api, router_name).execute(check_error=True)

        router = self.nb_api.lr_get(router_name).execute(check_error=True)
        self.assertEqual(router.name, router_name)

    def test_reconcile_router_updates_existing_options(self):
        router_name = _get_unique_name()

        self.nb_api.lr_add(
            router_name, options={'wrong-option': 'value'}).execute(
                check_error=True)

        commands.ReconcileRouterCommand(
            self.nb_api, router_name).execute(check_error=True)

        router = self.nb_api.lr_get(router_name).execute(check_error=True)
        self.assertEqual(router.name, router_name)


class ReconcileMainRouterCommandTestCase(bgp.BaseBgpNbIdlTestCase):
    def setUp(self):
        super().setUp()
        self.router_name = _get_unique_name()
        cfg.CONF.set_override('main_router_name', self.router_name, 'bgp')

    def _validate_main_router_options(self):
        router = self.nb_api.lr_get(self.router_name).execute(check_error=True)
        self.assertEqual(router.name, self.router_name)
        self.assertEqual(router.options.get('dynamic-routing'), 'true')
        self.assertEqual(router.options.get('dynamic-routing-redistribute'),
                         'connected-as-host,nat')

    def test_reconcile_main_router_with_dynamic_routing(self):
        commands.ReconcileMainRouterCommand(
            self.nb_api).execute(check_error=True)

        self._validate_main_router_options()

    def test_reconcile_updates_existing_main_router_options(self):
        self.nb_api.lr_add(
            self.router_name,
            options={'dynamic-routing': 'false', 'wrong-option': 'value'}
        ).execute(check_error=True)

        commands.ReconcileMainRouterCommand(
            self.nb_api).execute(check_error=True)

        self._validate_main_router_options()


class ReconcileChassisRouterCommandTestCase(bgp.BaseBgpNbIdlTestCase):
    def test_reconcile_chassis_router(self):
        chassis = _create_fake_chassis()
        router_name = helpers.get_chassis_router_name(chassis.name)

        commands.ReconcileChassisRouterCommand(
            self.nb_api, chassis).execute(check_error=True)

        router = self.nb_api.lr_get(router_name).execute(check_error=True)
        self.assertEqual(router.name, router_name)
        self.assertEqual(router.options.get('chassis'), chassis.name)

    def test_reconcile_updates_chassis_router_options(self):
        chassis = _create_fake_chassis()
        router_name = helpers.get_chassis_router_name(chassis.name)

        self.nb_api.lr_add(
            router_name,
            options={'chassis': 'wrong-chassis', 'other-option': 'value'}
        ).execute(check_error=True)

        commands.ReconcileChassisRouterCommand(
            self.nb_api, chassis).execute(check_error=True)

        router = self.nb_api.lr_get(router_name).execute(check_error=True)
        self.assertEqual(router.options.get('chassis'), chassis.name)


class ConnectChassisRouterToMainRouterCommandTestCase(
        bgp.BaseBgpNbIdlTestCase):
    def setUp(self):
        super().setUp()
        self.main_router_name = _get_unique_name()
        cfg.CONF.set_override('main_router_name', self.main_router_name, 'bgp')

        self.fake_chassis = _create_fake_chassis()
        self.chassis_router_name = helpers.get_chassis_router_name(
            self.fake_chassis.name)

        hcg_name = f'bgp-hcg-{self.fake_chassis.name}'
        self.hcg_id = self._create_hcg(hcg_name)

        commands.ReconcileMainRouterCommand(
            self.nb_api).execute(check_error=True)
        commands.ReconcileChassisRouterCommand(
            self.nb_api,
            self.fake_chassis).execute(check_error=True)

    def _create_hcg(self, hcg_name):
        return commands._HAChassisGroupAddCommand(
            self.nb_api, hcg_name).execute(check_error=True).uuid

    def _validate_connection_created(self):
        lrp_main_name = helpers.get_lrp_name(
            self.main_router_name, self.chassis_router_name)
        lrp_chassis_name = helpers.get_lrp_name(
            self.chassis_router_name, self.main_router_name)

        lrp_main = self.nb_api.lrp_get(lrp_main_name).execute(check_error=True)
        lrp_chassis = self.nb_api.lrp_get(
            lrp_chassis_name).execute(check_error=True)

        # Check ports are connected
        self.assertEqual(lrp_main.peer, [lrp_chassis_name])
        self.assertEqual(lrp_chassis.peer, [lrp_main_name])

        # Verify main router LRP has HA chassis group
        self.assertEqual(self.hcg_id, lrp_main.ha_chassis_group[0].uuid)

        # Verify main router LRP has dynamic routing option
        self.assertEqual(
            'true', lrp_main.options.get('dynamic-routing-maintain-vrf'))

        return lrp_main_name, lrp_chassis_name

    def test_connect_router_to_main_router_new(self):
        commands.ConnectChassisRouterToMainRouterCommand(
            self.nb_api, self.fake_chassis, self.hcg_id
        ).execute(check_error=True)

        self._validate_connection_created()

    def test_connect_existing_lrps_get_updated(self):
        lrp_main_name = helpers.get_lrp_name(
            self.main_router_name, self.chassis_router_name)
        lrp_chassis_name = helpers.get_lrp_name(
            self.chassis_router_name, self.main_router_name)

        self.nb_api.lrp_add(
            self.chassis_router_name, lrp_chassis_name,
            mac='00:00:00:00:00:01',  # wrong MAC
            networks=['10.0.0.1/24'],  # wrong IP
            peer='wrong-peer'  # wrong peer
        ).execute(check_error=True)

        self.nb_api.lrp_add(
            self.main_router_name, lrp_main_name,
            mac='00:00:00:00:00:02',  # wrong MAC
            networks=['10.0.0.2/24'],  # wrong IP
            peer='wrong-peer'  # wrong peer
        ).execute(check_error=True)

        commands.ConnectChassisRouterToMainRouterCommand(
            self.nb_api, self.fake_chassis, self.hcg_id
        ).execute(check_error=True)

        self._validate_connection_created()

    def test_connect_router_is_idempotent(self):
        commands.ConnectChassisRouterToMainRouterCommand(
            self.nb_api, self.fake_chassis, self.hcg_id
        ).execute(check_error=True)

        lrp_main1, lrp_chassis1 = self._validate_connection_created()

        commands.ConnectChassisRouterToMainRouterCommand(
            self.nb_api, self.fake_chassis, self.hcg_id
        ).execute(check_error=True)

        lrp_main2, lrp_chassis2 = self._validate_connection_created()

        self.assertEqual(lrp_main1, lrp_main2)
        self.assertEqual(lrp_chassis1, lrp_chassis2)

    def test_connect_router_to_non_existing_main_router(self):
        self.nb_api.lr_del(self.main_router_name).execute(check_error=True)
        cmd = commands.ConnectChassisRouterToMainRouterCommand(
            self.nb_api, self.fake_chassis, self.hcg_id)
        self.assertRaises(
            exceptions.ReconcileError,
            cmd.execute,
            check_error=True
        )

    def test_connect_non_existing_router_to_main_router(self):
        self.nb_api.lr_del(self.chassis_router_name).execute(check_error=True)
        cmd = commands.ConnectChassisRouterToMainRouterCommand(
            self.nb_api, self.fake_chassis, self.hcg_id)
        self.assertRaises(
            exceptions.ReconcileError,
            cmd.execute,
            check_error=True
        )


class ConnectRouterToSwitchCommandTestCase(bgp.BaseBgpNbIdlTestCase):
    def setUp(self):
        super().setUp()
        self.lr_name = _get_unique_name()
        self.ls_name = _get_unique_name()

        self.nb_api.lr_add(self.lr_name).execute(check_error=True)
        self.nb_api.ls_add(self.ls_name).execute(check_error=True)

    @bgp.requires_ovn_version_with_bgp()
    def test_connect_router_to_switch_without_ip(self):
        commands.ConnectRouterToSwitchCommand(
            self.nb_api, self.lr_name, self.ls_name).execute(check_error=True)

        lrp_name = helpers.get_lrp_name(self.lr_name, self.ls_name)
        lrp = self.nb_api.lrp_get(lrp_name).execute(check_error=True)
        self.assertEqual([], lrp.networks)

        lsp_name = helpers.get_lsp_name(self.ls_name, self.lr_name)
        lsp = self.nb_api.lsp_get(lsp_name).execute(check_error=True)
        self.assertEqual('router', lsp.type)
        self.assertEqual(['router'], lsp.addresses)
        self.assertEqual(lrp_name, lsp.options.get('router-port'))

    def test_connect_router_to_switch_with_ip(self):
        lrp_ip = '192.168.1.1/24'

        commands.ConnectRouterToSwitchCommand(
            self.nb_api, self.lr_name, self.ls_name, [lrp_ip]
        ).execute(check_error=True)

        lrp_name = helpers.get_lrp_name(self.lr_name, self.ls_name)
        lrp = self.nb_api.lrp_get(lrp_name).execute(check_error=True)
        self.assertEqual([lrp_ip], lrp.networks)

    @bgp.requires_ovn_version_with_bgp()
    def test_connect_existing_with_different_attributes(self):
        lrp_name = helpers.get_lrp_name(self.lr_name, self.ls_name)
        lsp_name = helpers.get_lsp_name(self.ls_name, self.lr_name)

        # Create LRP and LSP with wrong attributes
        self.nb_api.lrp_add(
            self.lr_name, lrp_name,
            mac='00:00:00:00:00:01',  # wrong MAC
            networks=['10.0.0.1/24']  # wrong networks
        ).execute(check_error=True)

        self.nb_api.lsp_add(
            self.ls_name, lsp_name,
            type='patch',  # wrong type
            addresses=['00:00:00:00:00:01'],  # wrong addresses
            options={'peer': 'wrong-peer'}
        ).execute(check_error=True)

        commands.ConnectRouterToSwitchCommand(
            self.nb_api, self.lr_name, self.ls_name).execute(check_error=True)

        lrp = self.nb_api.lrp_get(lrp_name).execute(check_error=True)
        self.assertEqual([], lrp.networks)

        lsp = self.nb_api.lsp_get(lsp_name).execute(check_error=True)
        self.assertEqual('router', lsp.type)
        self.assertEqual(['router'], lsp.addresses)
        self.assertEqual(lrp_name, lsp.options.get('router-port'))


class ConnectChassisRouterToSwitchCommandTestCase(bgp.BaseBgpNbIdlTestCase):
    def setUp(self):
        super().setUp()
        self.lr_name = _get_unique_name()
        self.ls_name = _get_unique_name()

        self.nb_api.lr_add(self.lr_name).execute(check_error=True)
        self.nb_api.ls_add(self.ls_name).execute(check_error=True)

    def test_sets_external_ids(self):
        network_name = _get_unique_name()
        commands.ConnectChassisRouterToSwitchCommand(
            self.nb_api, self.lr_name, self.ls_name, network_name
        ).execute(check_error=True)

        lrp_name = helpers.get_lrp_name(self.lr_name, self.ls_name)
        lrp = self.nb_api.lrp_get(lrp_name).execute(check_error=True)
        self.assertEqual(
            network_name,
            lrp.external_ids[constants.LRP_NETWORK_NAME_EXT_ID_KEY])


class ReconcileChassisCommandTestCase(bgp.BaseBgpTestCase):
    def setUp(self):
        super().setUp()
        self.main_router_name = _get_unique_name()
        cfg.CONF.set_override('main_router_name', self.main_router_name, 'bgp')

    def _create_chassis(self, name=None, ip=1, network_names=None):
        chassis_name = name or _get_unique_name("chassis")
        ext_ids = {}
        if network_names:
            ext_ids[constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY] = ','.join(
                network_names)
        return self.add_fake_chassis(
            chassis_name, f'172.24.4.{ip}',
            external_ids=ext_ids)

    def _validate_hcg_created(self, chassis_name):
        hcg_name = helpers.get_hcg_name(chassis_name)
        hcg = self.nb_api.db_find(
            'HA_Chassis_Group',
            ('name', '=', hcg_name)
        ).execute(check_error=True)
        self.assertTrue(hcg)
        return hcg[0]

    def _validate_chassis_router_created(self, chassis_name):
        router_name = helpers.get_chassis_router_name(chassis_name)
        router = self.nb_api.lr_get(router_name).execute(check_error=True)
        self.assertEqual(router_name, router.name)
        self.assertEqual(chassis_name, router.options.get('chassis'))
        return router

    def _validate_main_router_connection(self, chassis_name):
        chassis_router_name = helpers.get_chassis_router_name(chassis_name)

        lrp_main_name = helpers.get_lrp_name(
            self.main_router_name, chassis_router_name)
        lrp_chassis_name = helpers.get_lrp_name(
            chassis_router_name, self.main_router_name)

        lrp_main = self.nb_api.lrp_get(lrp_main_name).execute(check_error=True)
        lrp_chassis = self.nb_api.lrp_get(lrp_chassis_name).execute(
            check_error=True)

        self.assertEqual([lrp_chassis_name], lrp_main.peer)
        self.assertEqual([lrp_main_name], lrp_chassis.peer)

    def _validate_chassis_router_policies(self, chassis_name, router, peers):
        def _check_policy(policy, peer):
            peer_switch_name = helpers.get_chassis_peer_switch_name(
                chassis_name, peer)
            output_lrp = helpers.get_lrp_name(router.name,
                                              self.main_router_name)
            inport_lrp = helpers.get_lrp_name(router.name, peer_switch_name)
            return (policy.match == f'inport==\"{inport_lrp}\"' and
                    policy.action == 'reroute' and
                    policy.output_port[0].name == output_lrp)

        checked_policies = []
        for peer in peers:
            for policy in router.policies:
                if _check_policy(policy, peer):
                    checked_policies.append(policy)
                    break
            else:
                self.fail(f"Policy for peer {peer} not found")

        self.assertEqual(len(router.policies), len(checked_policies))

    @bgp.requires_ovn_version_with_bgp()
    def test_reconcile_chassis_basic(self):
        peers = ["bgp1", "bgp2"]
        chassis = self._create_chassis(network_names=peers)

        # Create main router first (prerequisite)
        commands.ReconcileMainRouterCommand(self.nb_api).execute(
            check_error=True)

        commands.ReconcileChassisCommand(
            self.nb_api, chassis).execute(check_error=True)

        router = self.nb_api.lr_get(
            helpers.get_chassis_router_name(chassis.name)).execute(
                check_error=True)
        # Validate all components were created
        self._validate_hcg_created(chassis.name)
        self._validate_chassis_router_created(chassis.name)
        self._validate_main_router_connection(chassis.name)
        self._validate_chassis_router_policies(chassis.name, router, peers)

    @bgp.requires_ovn_version_with_bgp()
    def test_reconcile_chassis_idempotent(self):
        peers = ["bgp1", "bgp2"]
        chassis = self._create_chassis(network_names=peers)

        commands.ReconcileMainRouterCommand(self.nb_api).execute(
            check_error=True)

        cmd = commands.ReconcileChassisCommand(self.nb_api, chassis)
        cmd.execute(check_error=True)
        # Run again to check idempotency
        cmd.execute(check_error=True)

        router = self.nb_api.lr_get(
            helpers.get_chassis_router_name(chassis.name)).execute(
                check_error=True)
        self._validate_hcg_created(chassis.name)
        self._validate_chassis_router_created(chassis.name)
        self._validate_main_router_connection(chassis.name)
        self._validate_chassis_router_policies(chassis.name, router, peers)

    def test_reconcile_chassis_with_existing_components(self):
        chassis = self._create_chassis()
        hcg_name = helpers.get_hcg_name(chassis.name)
        router_name = helpers.get_chassis_router_name(chassis.name)

        # Create main router first
        commands.ReconcileMainRouterCommand(self.nb_api).execute(
            check_error=True)

        # Pre-create HCG with different settings
        self.nb_api.ha_chassis_group_add(hcg_name).execute(check_error=True)

        # Pre-create router with wrong chassis
        self.nb_api.lr_add(
            router_name,
            options={'chassis': 'wrong-chassis'}
        ).execute(check_error=True)

        # Execute command should update existing components
        commands.ReconcileChassisCommand(
            self.nb_api, chassis).execute(check_error=True)

        # Validate components were updated correctly
        router = self.nb_api.lr_get(router_name).execute(check_error=True)
        self.assertEqual(chassis.name, router.options.get('chassis'))

    def test_reconcile_chassis_missing_main_router(self):
        chassis = self._create_chassis()

        cmd = commands.ReconcileChassisCommand(self.nb_api, chassis)
        self.assertRaises(
            exceptions.ReconcileError,
            cmd.execute,
            check_error=True
        )


class BgpNbIdlWithChassisBase(bgp.BaseBgpNbIdlTestCase):
    def _create_chassis_lrp(self):
        chassis_router_name = _get_unique_name('chassis-lr')
        lrp_main_name = helpers.get_lrp_name(
            self.main_router_name, chassis_router_name)
        lrp_chassis_name = helpers.get_lrp_name(
            chassis_router_name, self.main_router_name)

        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.lr_add(chassis_router_name))
            txn.add(self.nb_api.lrp_add(
                chassis_router_name, lrp_chassis_name,
                mac=helpers.get_mac_address_from_lrp_name(lrp_chassis_name),
                networks=[],
                peer=lrp_main_name,
            ))
            lrp = txn.add(self.nb_api.lrp_add(
                self.main_router_name, lrp_main_name,
                mac=helpers.get_mac_address_from_lrp_name(lrp_main_name),
                peer=lrp_chassis_name,
                networks=[],
                external_ids={
                    constants.BGP_LRP_TO_CHASSIS: chassis_router_name},
            ))

        return lrp.result

    def _assert_main_router_has_policies_for_interconnect(
            self, interconnect_switch_name, chassis_lrps):
        interconnect_lrp_name = helpers.get_lrp_name(
            self.main_router_name, interconnect_switch_name)
        policies = self.nb_api.lr_policy_list(self.main_router_name).execute(
            check_error=True)
        interconnect_policies = [
            p for p in policies
            if interconnect_lrp_name in p.match and p.priority == 10]
        self.assertEqual(
            len(chassis_lrps), len(interconnect_policies),
            "Expected %d policies for chassis LRPs, got %d" % (
                len(chassis_lrps), len(interconnect_policies)))
        for lrp in chassis_lrps:
            expected_match = (
                'inport=="%s" && is_chassis_resident("cr-%s")' % (
                    interconnect_lrp_name, lrp.name))
            matching = [p for p in interconnect_policies
                        if p.match == expected_match and p.action == 'reroute']
            self.assertEqual(
                1, len(matching),
                "Expected one policy for chassis LRP %s with match %r" % (
                    lrp.name, expected_match))


class _BaseNeutronSwitchCommandTestCase(BgpNbIdlWithChassisBase):
    Subnet = collections.namedtuple('Subnet', ['cidr', 'gateway_ip'])

    def setUp(self):
        super().setUp()
        self.main_router_name = bgp_config.get_main_router_name()
        commands.ReconcileMainRouterCommand(
            self.nb_api).execute(check_error=True)

    def _create_neutron_switch_with_localnet(self, network_name, subnets):
        """Simulate a Neutron provider network"""
        net_id = uuidutils.generate_uuid()
        ls_name = f'neutron-{net_id}'
        localnet_lsp_name = f'{ls_name}-localnet'

        with self.nb_api.transaction(check_error=True) as txn:
            ls = txn.add(self.nb_api.ls_add(ls_name))
            for subnet in subnets:
                txn.add(ovn_commands.AddDHCPOptionsCommand(
                    self.nb_api, net_id, may_exist=False,
                    cidr=subnet.cidr,
                    options={'router': subnet.gateway_ip},
                    external_ids={ovn_const.OVN_NETWORK_ID_EXT_ID_KEY: net_id})
                )
            txn.add(self.nb_api.lsp_add(
                ls_name, localnet_lsp_name,
                type='localnet',
                options={'network_name': network_name},
                addresses=['unknown']
            ))

        return ls.result


class ReconcileNeutronSwitchCommandTestCase(_BaseNeutronSwitchCommandTestCase):
    def _validate_neutron_switch_dead_connection(self, n_switch):
        lrp_name = helpers.get_lrp_name(self.main_router_name, n_switch.name)
        lsp_name = helpers.get_lsp_name(n_switch.name, self.main_router_name)

        lrp = self.nb_api.lrp_get(lrp_name).execute(check_error=True)
        self.assertIsNotNone(lrp)

        lsp = self.nb_api.lsp_get(lsp_name).execute(check_error=True)
        self.assertEqual('router', lsp.type)
        self.assertEqual(lrp_name, lsp.options.get('router-port'))

    def _validate_interconnect_switch_created(self, n_switch, network_name):
        interconnect_name = helpers.get_provider_interconnect_switch_name(
            n_switch.name)

        ls = self.nb_api.ls_get(interconnect_name).execute(check_error=True)
        self.assertEqual(interconnect_name, ls.name)

        localnet_lsp_name = helpers.get_lsp_localnet_name(interconnect_name)
        lsp = self.nb_api.lsp_get(localnet_lsp_name).execute(check_error=True)
        self.assertEqual('localnet', lsp.type)
        self.assertEqual(network_name, lsp.options.get('network_name'))

    def _validate_interconnect_switch_connection(self, n_switch, gw_ips):
        interconnect_name = helpers.get_provider_interconnect_switch_name(
            n_switch.name)
        lrp_name = helpers.get_lrp_name(
            self.main_router_name, interconnect_name)

        lrp = self.nb_api.lrp_get(lrp_name).execute(check_error=True)
        self.assertIsNotNone(lrp)

        # Check gateway IPs are set
        self.assertCountEqual(gw_ips, lrp.networks)

        return lrp

    @bgp.requires_ovn_version_with_bgp()
    def test_reconcile_neutron_switch_basic(self):
        network_name = 'provider-net'
        n_switch = self._create_neutron_switch_with_localnet(network_name, [])

        # Create 3 chassis routers
        chassis_lrps = [self._create_chassis_lrp() for _ in range(3)]

        commands.ReconcileNeutronSwitchCommand(
            self.nb_api, n_switch
        ).execute(check_error=True)

        self._validate_neutron_switch_dead_connection(n_switch)
        self._validate_interconnect_switch_created(n_switch, network_name)
        self._validate_interconnect_switch_connection(n_switch, gw_ips=[])
        interconnect_name = helpers.get_provider_interconnect_switch_name(
            n_switch.name)
        self._assert_main_router_has_policies_for_interconnect(
            interconnect_name, chassis_lrps)

    def test_reconcile_neutron_switch_with_gateway_ips(self):
        network_name = 'provider-net'
        subnets = [
            self.Subnet(cidr='192.168.1.1/24', gateway_ip='192.168.1.1'),
            self.Subnet(cidr='2001:db8::/64', gateway_ip='2001:db8::5'),
        ]
        n_switch = self._create_neutron_switch_with_localnet(
            network_name, subnets)

        commands.ReconcileNeutronSwitchCommand(
            self.nb_api, n_switch
        ).execute(check_error=True)

        self._validate_neutron_switch_dead_connection(n_switch)
        self._validate_interconnect_switch_created(n_switch, network_name)

        gw_ips = ['192.168.1.1/24', '2001:db8::5/64']
        lrp = self._validate_interconnect_switch_connection(n_switch, gw_ips)
        self.assertCountEqual(gw_ips, lrp.networks)
        interconnect_name = helpers.get_provider_interconnect_switch_name(
            n_switch.name)
        self._assert_main_router_has_policies_for_interconnect(
            interconnect_name, [])

    def test_reconcile_neutron_switch_idempotent(self):
        network_name = 'provider-net'
        subnets = [
            self.Subnet(cidr='192.168.1.0/24', gateway_ip='192.168.1.4'),
        ]
        n_switch = self._create_neutron_switch_with_localnet(
            network_name, subnets)

        commands.ReconcileNeutronSwitchCommand(
            self.nb_api, n_switch
        ).execute(check_error=True)
        commands.ReconcileNeutronSwitchCommand(
            self.nb_api, n_switch
        ).execute(check_error=True)

        self._validate_neutron_switch_dead_connection(n_switch)
        self._validate_interconnect_switch_created(n_switch, network_name)

        gw_ips = ['192.168.1.4/24']
        self._validate_interconnect_switch_connection(n_switch, gw_ips)
        interconnect_name = helpers.get_provider_interconnect_switch_name(
            n_switch.name)
        self._assert_main_router_has_policies_for_interconnect(
            interconnect_name, [])

    def test_reconcile_neutron_switch_no_localnet_port_raises(self):
        ls_name = _get_unique_name('neutron')
        self.nb_api.ls_add(ls_name).execute(check_error=True)
        n_switch = self.nb_api.ls_get(ls_name).execute(check_error=True)

        self.assertRaises(
            ValueError,
            commands.ReconcileNeutronSwitchCommand,
            self.nb_api, n_switch
        )


class ReconcileMainRouterPoliciesCommandTestCase(BgpNbIdlWithChassisBase):
    def setUp(self):
        super().setUp()
        self.main_router_name = bgp_config.get_main_router_name()
        commands.ReconcileMainRouterCommand(
            self.nb_api).execute(check_error=True)
        self.main_router = self.nb_api.lr_get(
            self.main_router_name).execute(check_error=True)

    def _create_interconnect_switch(self, gw_ips=None):
        gw_ips = gw_ips or []
        ls_name = _get_unique_name('interconnect')
        lrp_name = helpers.get_lrp_name(self.main_router_name, ls_name)
        lsp_name = helpers.get_lsp_name(ls_name, self.main_router_name)

        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.ls_add(ls_name))
            txn.add(self.nb_api.lrp_add(
                self.main_router_name, lrp_name,
                mac=helpers.get_mac_address_from_lrp_name(lrp_name),
                networks=gw_ips,
            ))
            txn.add(self.nb_api.lsp_add(
                ls_name, lsp_name,
                type='router',
                addresses=['router'],
                options={'router-port': lrp_name},
            ))

        return ls_name, lrp_name

    @bgp.requires_ovn_version_with_bgp()
    def test_creates_policy_and_route(self):
        chassis_lrp = self._create_chassis_lrp()
        interconnect_switch_name, interconnect_lrp_name = (
            self._create_interconnect_switch())

        commands.ReconcileMainRouterPoliciesCommand(
            self.nb_api,
            self.main_router,
            interconnect_lrp_name,
            chassis_lrp,
        ).execute(check_error=True)

        self._assert_main_router_has_policies_for_interconnect(
            interconnect_switch_name, [chassis_lrp])

    @bgp.requires_ovn_version_with_bgp()
    def test_creates_policy_and_route_with_related_resource(self):
        chassis_lrp = self._create_chassis_lrp()
        interconnect_name, interconnect_lrp_name = (
            self._create_interconnect_switch())

        related_switch = self.nb_api.ls_get(interconnect_name).execute(
            check_error=True)

        commands.ReconcileMainRouterPoliciesCommand(
            self.nb_api,
            self.main_router,
            interconnect_lrp_name,
            chassis_lrp,
            related_resource=related_switch,
        ).execute(check_error=True)

        self._assert_main_router_has_policies_for_interconnect(
            interconnect_name, [chassis_lrp])

    @bgp.requires_ovn_version_with_bgp()
    def test_idempotent(self):
        chassis_lrp = self._create_chassis_lrp()
        interconnect_switch_name, interconnect_lrp_name = (
            self._create_interconnect_switch())

        for _ in range(2):
            commands.ReconcileMainRouterPoliciesCommand(
                self.nb_api,
                self.main_router,
                interconnect_lrp_name,
                chassis_lrp,
            ).execute(check_error=True)

        self._assert_main_router_has_policies_for_interconnect(
            interconnect_switch_name, [chassis_lrp])


class ReconcileMainRouterRoutesForProviderCommandTestCase(
        BgpNbIdlWithChassisBase):

    def setUp(self):
        super().setUp()
        self.main_router_name = bgp_config.get_main_router_name()
        commands.ReconcileMainRouterCommand(
            self.nb_api).execute(check_error=True)

    def _create_interconnect_switch(self, name=None):
        ls_name = name or _get_unique_name('interconnect')

        lrp_name = helpers.get_lrp_name(self.main_router_name, ls_name)

        with self.nb_api.transaction(check_error=True) as txn:
            ls = txn.add(self.nb_api.ls_add(ls_name))
            txn.add(self.nb_api.lrp_add(
                self.main_router_name, lrp_name,
                mac=helpers.get_mac_address_from_lrp_name(lrp_name),
                networks=['192.168.1.1/24'],
            ))

        return ls.result

    @bgp.requires_ovn_version_with_bgp()
    def test_creates_policies_for_each_chassis_lrp(self):
        chassis_lrp1 = self._create_chassis_lrp()
        chassis_lrp2 = self._create_chassis_lrp()
        interconnect_switch = self._create_interconnect_switch()

        commands.ReconcileMainRouterPoliciesForProviderCommand(
            self.nb_api,
            interconnect_switch.name,
            related_resource=interconnect_switch,
        ).execute(check_error=True)

        self._assert_main_router_has_policies_for_interconnect(
            interconnect_switch.name, [chassis_lrp1, chassis_lrp2])

    @bgp.requires_ovn_version_with_bgp()
    def test_no_chassis_lrps_creates_nothing(self):
        interconnect_switch = self._create_interconnect_switch()

        commands.ReconcileMainRouterPoliciesForProviderCommand(
            self.nb_api,
            interconnect_switch.name,
            related_resource=interconnect_switch,
        ).execute(check_error=True)

        self._assert_main_router_has_policies_for_interconnect(
            interconnect_switch.name, [])

        router = self.nb_api.lr_get(self.main_router_name).execute(
            check_error=True)
        routes = [r for r in router.static_routes
                  if r.external_ids.get(
                    constants.RELATED_RESOURCE_TAG) ==
                    str(interconnect_switch.uuid)]
        self.assertEqual(0, len(routes))

    @bgp.requires_ovn_version_with_bgp()
    def test_sets_related_resource_tag(self):
        chassis_lrp = self._create_chassis_lrp()
        interconnect_switch = self._create_interconnect_switch()

        commands.ReconcileMainRouterPoliciesForProviderCommand(
            self.nb_api,
            interconnect_switch.name,
            related_resource=interconnect_switch,
        ).execute(check_error=True)

        self._assert_main_router_has_policies_for_interconnect(
            interconnect_switch.name, [chassis_lrp])



# Copyright (c) 2024 Red Hat Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from neutron_lib.api.definitions import uplink_status_propagation as usp_def
from neutron_lib.api.definitions import uplink_status_propagation_updatable \
    as uspu_def
from neutron_lib.plugins import directory
from oslo_config import cfg

from neutron.tests.unit.plugins.ml2 import test_plugin


@ddt.ddt
class UplinkStatusPropagationUpdatableML2ExtDriverTestCase(
        test_plugin.Ml2PluginV2TestCase):

    _extension_drivers = [usp_def.ALIAS.replace('-', '_'),
                          uspu_def.ALIAS.replace('-', '_')]

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                              self._extension_drivers,
                              group='ml2')
        super().setUp()
        self.plugin = directory.get_plugin()

    @ddt.data(True, False)
    def test_port_update_propagate_uplink_status(self, _status):
        with self.network() as n:
            args = {'port': {'name': 'test',
                             'network_id': n['network']['id'],
                             'project_id': n['network']['project_id'],
                             'device_id': '',
                             'device_owner': '',
                             'fixed_ips': '',
                             'propagate_uplink_status': _status,
                             'admin_state_up': True,
                             'status': 'ACTIVE'}}
            try:
                port = self.plugin.create_port(self.context, args)
                args = {'port': {'propagate_uplink_status': not _status}}
                self.plugin.update_port(self.context, port['id'], args)
                port = self.plugin.get_port(self.context, port['id'])
                self.assertEqual(not _status, port['propagate_uplink_status'])
            finally:
                if port:
                    self.plugin.delete_port(self.context, port['id'])
